# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# Making this explicit is often much easier to read. Stupid warning.
# pylint: disable=no-else-return

"""Miscellaneous small functions that don't need to be class methods."""

from __future__ import unicode_literals
from __future__ import print_function

import atexit
import codecs
import copy
import datetime
import json
import logging
import os
import re
import shutil
import tempfile
# pylint:disable=import-error
from six.moves.urllib.parse import urlparse, urlencode
from six.moves.urllib.request import urlopen, Request
from six.moves.urllib.error import URLError, HTTPError
from productmd.images import (SUPPORTED_IMAGE_TYPES, SUPPORTED_IMAGE_FORMATS)
import productmd.composeinfo

import fedfind.const

# pylint:disable=invalid-name
logger = logging.getLogger(__name__)

# Cache handling stuff
CACHEDIR = os.path.join(os.path.expanduser('~'), '.cache', 'fedfind')
try:
    if not os.path.exists(CACHEDIR):
        os.makedirs(CACHEDIR)
    if not os.access(CACHEDIR, os.W_OK):
        raise OSError("Write access to {0} denied!".format(CACHEDIR))
except (IOError, OSError) as err:
    logger.warning("Cannot use permanent cache! Using temporary cache, valid only for process")
    logger.warning("Error: %s", err)
    CACHEDIR = tempfile.mkdtemp()
    @atexit.register
    def _clean_cache():
        """Clean up the temporary cache."""
        logger.debug("Removing temporary cache %s", CACHEDIR)
        shutil.rmtree(CACHEDIR, True)

# cache for pkgdb collections JSON, as there are cases where we want
# to re-use it
COLLECTIONS = None

def date_check(value, fail_raise=True, out='obj'):
    """Checks whether a value is a date, and returns it if so. Valid
    dates are datetime.date instances or a string or int 'YYYYMMDD'.
    If out is set to 'obj', returns a datetime.date object. If out is
    set to 'both', returns a (string, object) tuple. If out is set to
    'str' (or anything else), returns a 'YYYYMMDD' string.

    On failure - empty or non-date-y input - will raise an exception
    if fail_raise is set, otherwise will return False.
    """
    if isinstance(value, datetime.date):
        dateobj = value
    else:
        try:
            dateobj = datetime.datetime.strptime(str(value), '%Y%m%d')
        except ValueError:
            if fail_raise:
                newerr = "{0} is not a valid date.".format(value)
                raise ValueError(newerr)
            else:
                return False

    if out == 'obj':
        return dateobj
    else:
        datestr = dateobj.strftime('%Y%m%d')
        if out == 'both':
            return (datestr, dateobj)
        else:
            return datestr

def urlopen_retries(url, tries=5):
    """Simple attempt to open a URL and return the response, with
    retries.
    """
    tries = 5
    resp = None
    while not resp and tries:
        try:
            resp = urlopen(url)
        except (ValueError, URLError, HTTPError) as err:
            logger.debug("HTTP error! Retrying...")
            logger.debug("Error: %s", err)
            tries -= 1
    if not resp:
        raise ValueError("urlopen_retries: Failed to open {0}!".format(url))
    return resp

def download_json(url):
    """Given a URL that should give us some JSON, download and return
    it.
    """
    resp = urlopen_retries(url)
    # We can't just use json.load on the response because of a bytes/
    # str type issue in Py3: https://stackoverflow.com/questions/6862770
    rawjson = resp.read().decode('utf8')
    return json.loads(rawjson)

def url_exists(url):
    """Checks whether a URL exists, by trying to open it."""
    scheme = urlparse(url).scheme
    if 'http' in scheme:
        logger.debug("url_exists: checking %s as HTTP", url)
        try:
            urlopen(url)
            return True
        except (ValueError, URLError, HTTPError):
            return False
    else:
        raise ValueError("Invalid or unhandled URL! URL: {0}".format(url))

def get_size(url):
    """Get the size of URL (Content-Length header for HTTP). Currently
    only supports HTTP.
    """
    logger.debug("Checking size of %s", url)
    if not url.startswith('http'):
        raise ValueError("Can only check HTTP URLs.")
    headers = urlopen_retries(url).info()
    size = headers.get('Content-Length')
    if not size:
        raise ValueError("Size could not be found!")
    return int(size)

def comma_list(string):
    """Split a comma-separated list of values into a list (used by the
    CLI for passing query parameters etc). Also lower-cases the string
    because it's convenient to do that here...
    """
    return string.lower().split(',')

def _get_collections():
    """Download the collections JSON if we haven't already, and cache
    it at the module level.
    """
    # yes, pylint, we really want a global. sheesh.
    # pylint:disable=global-statement
    global COLLECTIONS
    if not COLLECTIONS:
        logger.debug("collections not cached, downloading...")
        gotjson = download_json('https://admin.fedoraproject.org/pkgdb/api/collections/')
        COLLECTIONS = gotjson['collections']
    return COLLECTIONS

def get_current_release(branched=False):
    """Find out what the 'current' Fedora release is, according to
    the pkgdb API. If 'branched' is truth-y, will return the Branched
    release number if there is one.
    """
    collections = _get_collections()
    if branched:
        rels = [coll['version'] for coll in collections if
                coll['status'] in ('Active', 'Under Development')]
    else:
        rels = [coll['version'] for coll in collections if coll['status'] == 'Active']
    rels = [int(rel) for rel in rels if rel.isdigit()]
    return max(rels)

def get_current_stables():
    """Returns a list of the release numbers of all current stable
    releases (according to pkgdb).
    """
    collections = _get_collections()
    return [int(coll['version']) for coll in collections if
            coll['status'] == 'Active' and coll['name'] == 'Fedora']

def create_image_dict(path):
    # pylint:disable=too-many-branches
    """Synthesize a productmd-style image metadata dict by analyzing
    the image's path. Used for pre-Pungi 4 releases, and Pungi 4
    releases that were split and had metadata stripped when synced to
    mirrors.
    """
    logger.debug("Synthesizing image metadata for %s", path)
    imgdict = {'path': path.strip('/')}
    path = path.lower()
    filename = path.split('/')[-1]

    # Find arch
    imgdict['arch'] = ""
    for arc in fedfind.const.ARCHES:
        if any(val in path for val in arc.values):
            imgdict['arch'] = arc.name

    # Find format
    imgdict['format'] = ""
    for form in SUPPORTED_IMAGE_FORMATS:
        if filename.endswith(form):
            imgdict['format'] = form

    # Find type
    imgdict['type'] = ""
    # we invent a types for the 'multiple' images that contain
    # multiple live or installer images together with a boot menu.
    # these will have 'Multi' subvariant
    types = SUPPORTED_IMAGE_TYPES + ['multi-desktop', 'multi-install']
    for typ in types:
        # Some types are basically formats...
        if '-' in typ and typ.replace('-', '.') in filename:
            imgdict['type'] = typ
        # Others are more token-y.
        elif "-{0}-".format(typ) in filename:
            imgdict['type'] = typ
    # Sometimes we have to be a bit more relaxed.
    if not imgdict['type']:
        for typ in types:
            if typ != "cd" and typ in filename:
                imgdict['type'] = typ
    # and this is just magic.
    if not imgdict['type'] and "disc" in filename:
        imgdict['type'] = "cd"
    # Pungi config uses the type 'boot' for netinst images, so follow
    # that
    if imgdict['type'] == 'netinst':
        imgdict['type'] = 'boot'
    # Various special handlings for awkward ostree installer images
    # Since 2016-10-10 with up-to-date Pungi the filenames should have
    # -ostree- in them...
    if '-ostree-' in filename:
        imgdict['type'] = "dvd-ostree"
    # ...but as of 2016-10-08 F24 two-week Atomic composes, the
    # filename is e.g.: Fedora-Atomic-dvd-x86_64-24-20161008.0.iso
    if '-atomic-dvd-' in filename:
        imgdict['type'] = "dvd-ostree"
    # F23 release case: Fedora-Cloud_Atomic-x86_64-23.iso
    if imgdict['type'] in ("", "dvd") and "cloud_atomic" in filename:
        imgdict['type'] = "dvd-ostree"
    # awkward case from F25 Beta, where we got a Workstation atomic
    # installer image with a bad filename:
    # Fedora-Workstation-dvd-x86_64-25_Beta-1.1.iso
    # probably too broad to keep forever, drop when 25 Beta is gone
    if '-workstation-dvd-' in filename:
        imgdict['type'] = "dvd-ostree"
    # live respins don't have 'live' in the filenames, find them with
    # the regex
    if not imgdict['type'] and fedfind.const.RESPINRE.search(path):
        imgdict['type'] = "live"

    # Find subvariant
    imgdict['subvariant'] = ''
    # these are a couple of nasty special cases taken
    # from fedfind 1.x
    if re.match(r'.*Fedora-(x86_64|i386)-\d{2,2}-\d{8,8}(\.\d|)-sda\.(qcow2|raw.xz)',
                imgdict['path']):
        # F19/20-era cloud image, before Cloud flavor.
        imgdict['subvariant'] = 'Cloud'

    if re.match(r'.*(Fedora-|F)\d{1,2}-(x86_64|i686|Live)-(x86_64|i686|Live)\.iso',
                imgdict['path']):
        # < F14 Desktop lives. Ew.
        imgdict['subvariant'] = 'Desktop'

    for load in fedfind.const.SUBVARIANTS:
        if any(val in imgdict['path'].lower() for val in load.values):
            imgdict['subvariant'] = load.name
    # Try to identify 'generic' media (call them
    # 'Everything' as that's what productmd does)
    if not imgdict['subvariant'] and (imgdict['type'] in ('dvd', 'boot', 'cd')):
        imgdict['subvariant'] = 'Everything'

    # assign disc number
    # FIXME: can't do disc_count very well with this approach
    imgdict['disc_number'] = 1
    match = re.search(r'-disc(\d+)', filename)
    if match:
        imgdict['disc_number'] = int(match.group(1))

    return imgdict

def pdc_query(path, params=None):
    # pylint:disable=too-many-locals,too-many-statements
    """Light 'API client', sorta, for PDC. Just handles sending a
    request to a given API path with the appropriate header to get
    a JSON response, parsing the response, and handling multi-page
    responses. params is a dict of query params. Caches certain
    responses which are very unlikely ever to change in fedfind's
    cache directory.
    """
    logger.debug("pdc_query: path is %s", path)
    # load the query cache
    cachefile = os.path.join(CACHEDIR, 'pdcquery.json')
    try:
        with codecs.open(cachefile, 'r', encoding='utf-8') as cachefh:
            cache = json.loads(cachefh.read())
    except (IOError, OSError):
        # file doesn't exist, etc.
        cache = {}

    # handle params (as we use them in the cache key)
    if not params:
        params = []
    # turn param dict into tuple list (we need to handle both dicts
    # and tuple lists, and you can't always turn a tuple list into a
    # dict because parameter names can be repeated)
    try:
        params = list(params.items())
    except AttributeError:
        pass
    baseurl = '/'.join((fedfind.const.PDC_API, path.strip('/'))) + '/'
    logger.debug("pdc_query: baseurl is: %s", baseurl)
    # cache key is the URL of the query, but without the page param
    cachekey = '?'.join((baseurl, urlencode(params)))

    # check cache, return if hit
    try:
        ret = cache[cachekey]
        logger.debug("pdc_query: cache hit!")
        return ret
    except KeyError:
        pass

    # otherwise, actually do the query
    results = []
    # we figured out the params earlier, for the cache key, just add
    # 'page' and re-do the URL
    pgcnt = 1
    params.append(('page', str(pgcnt)))
    url = '?'.join((baseurl, urlencode(params)))
    headers = {'Content-type': 'application/json'}
    req = Request(url, headers=headers)
    nxt = True
    while nxt:
        resp = download_json(req)
        try:
            # paged query
            results.extend(resp['results'])
            params.remove(('page', str(pgcnt)))
            pgcnt += 1
            params.append(('page', str(pgcnt)))
            url = '?'.join((baseurl, urlencode(params)))
            req = Request(url, headers=headers)
            nxt = resp['next']
        except KeyError:
            # non-paged query. just take the response
            results = resp
            break

    # decide whether to cache results
    writecache = False
    # we always cache compose-images results, as if we get a result
    # it's a 'hit' and will never change (query for non-existent
    # compose gives a 404)
    if path.startswith('compose-images/'):
        logger.debug("pdc_query: caching result for %s", path)
        cache[cachekey] = results
        writecache = True
    # we cache 'composes' results if they're specific and we got a
    # result
    if path == 'composes' and results:
        parkeys = [param[0] for param in params]
        if 'compose_id' in parkeys or all(key in parkeys for key in ('release', 'compose_label')):
            logger.debug("pdc_query: caching 'composes' result")
            cache[cachekey] = results
            writecache = True

    # write cache if we decided to
    if writecache:
        # I've seen CACHEDIR suddenly vanish, when it's a temp dir; let's
        # check and re-create it if necessary
        if not os.path.exists(CACHEDIR):
            os.makedirs(CACHEDIR)

        with codecs.open(cachefile, 'w', encoding='utf-8') as cachefh:
            cachefh.write(json.dumps(cache, separators=(',', ':'), indent=0))

    return results

def find_cid(string):
    """Find a Fedora compose ID (or a fedfind fake CID) in a string."""
    # the 'rules' are really pretty lax, see productmd.composeinfo
    # _validate_id's regex:
    # r".*\d{8}(\.nightly|\.n|\.ci|\.test|\.t)?(\.\d+)?"
    # it just has to be a string with 8 digits in it somewhere. We are
    # stricter, though, for Fedora. We require:
    # (dist)-(version)-(date)<.type>.(respin)
    # the type is optional, but all others must be present. The dist
    # must start with 'Fedora', but can have more text after that (and
    # can contain dashes, because life is awful; in the rest of the ID
    # dashes can only be separators).
    cidpatt = re.compile(r'(Fedora|FACD).*-[^-]+-\d{8,8}(\.\w)?\.\d+')
    match = cidpatt.search(string)
    if match:
        return match.group(0)
    else:
        return ''

def parse_cid(origcid, dist=False, dic=False):
    """If dist is false-y, get (release, date, type, respin) values
    from a Pungi 4 compose id. If dist is truth-y, get (dist, release,
    date, type, respin) values from a Pungi 4 compose id.
    """
    # again, this is just a complex operation.
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    # Normalize and check cid
    cid = find_cid(origcid)
    if not cid:
        raise ValueError("{0} does not appear to be a valid Pungi 4 Fedora "
                         "compose ID!".format(origcid))
    # Writing a truly generic compose ID parser is pretty tricky. See
    # https://github.com/release-engineering/productmd/issues/77
    # for some background. This is my best effort. I'm trying to get
    # it upstreamed currently.

    # init values
    mainshort = ''
    mainver = ''
    maintype = ''
    baseshort = ''
    basever = ''
    basetype = ''
    variant = ''

    # find date, type, respin
    (date, typ, respin) = productmd.composeinfo.get_date_type_respin(cid)
    # now split on the date, we only care about what comes before
    part = cid.split(date)[0][:-1]

    # Handle "HACK: there are 2 RHEL 5 composes"
    if part.endswith("-Client"):
        variant = "Client"
        part = part[:-len('-Client')]
    elif part.endswith("-Server"):
        variant = "Server"
        part = part[:-len('-Server')]

    # Next part back must be either a version type suffix or a version
    # we don't know yet if this is the main version or the base
    # version for a layered product
    sometype = ''
    # copy the list of release types and sort it in descending order
    # by length, so we won't match on a subset of the entire type.
    _reltypes = list(productmd.common.RELEASE_TYPES)
    # updates-testing was only added in productmd 1.9, which is only
    # in Rawhide and F27 u-t at time of writing (2017-10-27). This can
    # be removed when 1.9 is stable in all releases
    if 'updates-testing' not in _reltypes:
        _reltypes.append('updates-testing')
    _reltypes.sort(key=len, reverse=True)
    for _reltype in _reltypes:
        # if part ends in _reltype, take it as 'sometype' and cut it
        if part.lower().endswith(_reltype.lower()):
            sometype = _reltype
            part = part[:-len(_reltype)-1]
            break
    # now whatever's at the end of 'part' must be a version
    somever = part.split('-')[-1]
    # strip it off
    part = '-'.join(part.split('-')[:-1])

    # what remains is either:
    # mainshort
    # or:
    # mainshort-mainver(-maintype)-baseshort
    # But this is where things get fun, because sadly, both mainshort
    # and baseshort could have - in them and mainver could have a type
    # suffix. So, life is fun. Working in our favour, shortnames are
    # *not* allowed to contain digits. Let's see if we can spot what
    # looks like a '-mainver'- component...
    elems = part.split('-')
    for (idx, cand) in enumerate(elems):
        # can't be the first or the last
        if idx == 0 or idx == len(elems) - 1:
            continue
        # now see if the cand looks like a version
        # FIXME: we should use RELEASE_VERSION_RE here, but not until
        # it enforces this stricter rule. We have decided that only
        # 'Rawhide' or anything that looks like a version number -
        # starts with a digit, can have any number of subsequent
        # digits and/or digit groups separated with single periods -
        # is a valid version.
        verreg = re.compile(r'(Rawhide|Bikeshed|[0-9]+(\.?[0-9]+)*)$')
        match = verreg.match(cand)
        if match:
            mainver = match.group(1)
            # check for a type suffix. of course, because life is
            # awful, some of these have dashes in them, so we get to
            # be careful...
            for _reltype in _reltypes:
                rtelems = _reltype.lower().split('-')
                if elems[idx+1:idx+1+len(rtelems)] == rtelems:
                    maintype = _reltype
                    # pop out the right number of elems
                    for _ in range(0, len(rtelems)):
                        elems.pop(idx+1)
                    break
            basever = somever
            basetype = sometype
            mainshort = '-'.join(elems[:idx])
            baseshort = '-'.join(elems[idx+1:])
            break

    # if we didn't establish a mainver above, we must not be layered,
    # and what remains is just mainshort, and somever is mainver
    if not mainshort:
        mainshort = '-'.join(elems)
        mainver = somever
        maintype = sometype

    if not maintype:
        maintype = 'ga'
    if basever and not basetype:
        basetype = 'ga'

    if dic:
        return {
            'short': mainshort,
            'version': mainver,
            'version_type': maintype,
            'base_short': baseshort,
            'base_version': basever,
            'base_type': basetype,
            'variant': variant,
            'date': date,
            'compose_type': typ,
            'respin': respin
        }
    elif dist:
        return (mainshort, mainver.lower(), date, typ, respin)
    else:
        return (mainver.lower(), date, typ, respin)

def cid_from_label(release, label, short='fedora'):
    """Get the compose ID for a compose by label. Must also specify
    the release as labels are re-used per release. Can specify a
    shortname, because not just 'Fedora' composes have labels; the
    stable release nightly composes (Cloud, Atomic, Docker) also do.
    We can't tell from the label itself what the shortname might be,
    in fact each day we should get three composes with identical
    labels, so there's really no option but to require the user to
    specify. Uses PDC.
    """
    short = short.lower()
    params = {'release': "{0}-{1}".format(short, release), 'compose_label': label}
    res = pdc_query('composes', params)
    if res:
        return res[0]['compose_id']
    return ''

def label_from_cid(cid):
    """Get the compose label for a compose by ID. Only completed
    production composes have discoverable labels; will return the
    empty string for other composes.
    """
    typ = parse_cid(cid, dic=True)['compose_type']
    if typ != 'production':
        return ''
    params = {'compose_id': cid}
    res = pdc_query('composes', params)
    if res:
        return res[0]['compose_label']
    return ''

def get_weight(imgdict, arch=True):
    """Given a productmd image dict, return a number that tries to
    represent how 'important' that image is. This is intended for use
    in producing things like download tables, so you can order the
    images sensibly. Used for e.g. wikitcms download tables. Consider
    arch if arch is truth-y, otherwise don't.
    """
    archscores = (
        (('x86_64', 'i386'), 2000),
    )
    loadscores = (
        (('everything',), 300),
        (('workstation',), 220),
        (('server',), 210),
        (('cloud', 'desktop', 'cloud_base', 'docker_base', 'atomic'), 200),
        (('kde',), 190),
        (('minimal',), 90),
        (('xfce',), 80),
        (('soas',), 73),
        (('mate',), 72),
        (('cinnamon',), 71),
        (('lxde',), 70),
        (('source',), -10),
    )
    imgscore = 0
    if arch:
        for (values, score) in archscores:
            if imgdict['arch'] in values:
                imgscore += score
    for (values, score) in loadscores:
        if imgdict['subvariant'].lower() in values:
            imgscore += score
    return imgscore

def correct_image(imgdict):
    """This function is intended to make 'corrections' to the image
    dict to handle cases where the metadata produced by pungi are
    problematic. The passed dict is copied and the modified dict
    returned.
    """
    newdict = copy.deepcopy(imgdict)
    # ostree installer images get type 'boot', but this is unviable
    # because network install images also get type 'boot' and thus
    # we cannot distinguish between an ostree installer image and a
    # network install image for the same subvariant, and since
    # 2016-10, Fedora has both a Workstation network install image
    # and a Workstation ostree installer image. This is filed as
    # https://pagure.io/pungi/issue/417 . Until it gets fixed, we'll
    # work around it here, by substituting the proposed 'dvd-ostree'
    # type value.
    if imgdict['type'] == 'boot':
        filename = imgdict['path'].split('/')[-1]
        if '-dvd-' in filename or '-ostree-' in filename:
            newdict['type'] = 'dvd-ostree'
    return newdict

def identify_image(imgdict, out='tuple', undersub=False, lower=False):
    """Produce an 'image identifier' from the image dict. We use the
    combination of subvariant, type, and format to 'identify' a given
    image all over the place; by having them all use this function we
    can make sure they're all consistent, handle special cases and
    ease changing the identifier in future if necessary. If 'out' is
    'string', you get a single string joined with dashes. If 'out' is
    'tuple', you get a tuple. If 'undersub' is true, dashes in the
    metadata values will be replaced by underscores (this is so the
    string form can be reliably split into the component parts). If
    lower is True, values will be forced to all lower case.
    """
    # 'correct' the imgdict first
    imgdict = correct_image(imgdict)

    # get the values
    subv = imgdict['subvariant']
    typ = imgdict['type']
    form = imgdict['format']

    if undersub:
        # sub - to _
        subv = subv.replace('-', '_')
        typ = typ.replace('-', '_')
        form = form.replace('-', '_')

    if lower:
        # lowercase
        subv = subv.lower()
        typ = typ.lower()
        form = form.lower()

    # construct the tuple
    ret = (subv, typ, form)

    if out == 'string':
        # produce the string
        ret = '-'.join(ret)

    return ret

def _get_filelist(flist='fedora', record=None):
    """Get a (read mode) filehandle for one of the create-filelist
    generated image lists from the mirrors. flist can be 'fedora',
    'alt' or 'archive'. record, if set, should be a dict. It will be
    given 'cache' and 'downfail' keys that are set to True or False
    depending on whether the cached file was returned, and whether the
    remote file download failed. This is really just for the use of
    the tests and shouldn't be considered part of the API.
    May raise OSError or IOError if cache file cannot be written, or
    ValueError if imagelist download fails and we don't have a cached
    copy. There's a bit of hidden magic here where we write the
    last-modified string from the request header as the first line of
    the cached file, then always iterate it out before returning the
    file handle.
    """
    # again, this is just a complex operation.
    # pylint: disable=too-many-branches,too-many-statements
    if record is not None:
        record['cache'] = False
        record['downfail'] = False
    # Determine the URL
    if flist not in ('fedora', 'alt', 'archive'):
        raise ValueError("_get_filelist: must be fedora, archive or alt!")
    fname = 'imagelist-{0}'.format(flist)
    url = '/'.join((fedfind.const.HTTPS_DL, flist, fname))
    logger.debug("_get_filelist: url is %s", url)

    # Determine the cache location
    cachefile = os.path.join(CACHEDIR, fname)
    logger.debug("_get_filelist: cachefile is %s", cachefile)

    # Get the cache file handle, if file is present
    try:
        cachefh = codecs.open(cachefile, 'r', encoding='utf-8')
    except (IOError, OSError):
        cachefh = None

    # Get the response and last modified date
    try:
        resp = urlopen_retries(url)
        modified = resp.headers['last-modified']
    except ValueError:
        if cachefh:
            logger.warning("Could not access image list {0}! Using cached "
                           "copy. Cache may be outdated!")
            if record:
                record['downfail'] = True
                record['cache'] = True
            # iterate out the date line
            next(cachefh)
            return cachefh
        else:
            # We can't do anything useful, just raise the error
            raise

    # Check if our cache is up-to-date, if present. If it is, go ahead
    # and return the file handle, with the date line already gone
    if cachefh:
        try:
            cachedate = next(cachefh).strip()
        except StopIteration:
            # file is empty
            cachedate = ''
        if cachedate == modified:
            logger.debug("_get_filelist: cache hit, returning")
            if record:
                logger.debug("Updating record")
                record['cache'] = True
            return cachefh

    # If we got here, either we don't have a cache file at all or the
    # dates don't match. So we need to re-download it

    # I've seen CACHEDIR suddenly vanish, when it's a temp dir; let's
    # check and re-create it if necessary
    if not os.path.exists(CACHEDIR):
        os.makedirs(CACHEDIR)

    with codecs.open(cachefile + '.tmp', 'w', encoding='utf-8') as cachefh:
        # Write the modified date line
        cachefh.write(modified + '\n')
    with open(cachefile + '.tmp', 'ab') as cachefh:
        # Now download and write the file, in 1MB chunks at a time
        while True:
            chunk = resp.read(1024 * 1024)
            if chunk:
                cachefh.write(chunk)
                cachefh.flush()
            else:
                break

    os.rename(cachefile + '.tmp', cachefile)

    # Now open the read-only filehandle, iterate out the date line,
    # and return
    cachefh = codecs.open(cachefile, 'r', encoding='utf-8')
    next(cachefh)
    return cachefh

# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2016 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=old-style-class, no-init, protected-access, no-self-use, unused-argument
# pylint: disable=too-many-public-methods

"""Tests for helpers.py."""

from __future__ import unicode_literals
from __future__ import print_function

import codecs
import datetime
import hashlib
import os
import shutil

import mock
import pytest
# pylint: disable=import-error
from six.moves.urllib.request import Request

import fedfind.const
import fedfind.helpers

URL = 'https://dl.fedoraproject.org'
PDC_JSON = {
    "results": ['some', 'results'],
    "next": False
}
PDC_JSON_REAL = {
    "results": [
        {
            'compose_id': 'Fedora-24-20160314.1',
            'compose_label': 'Alpha-1.1'
        }
    ],
    "next": False
}

SERVDVD_DICT = {
    "arch": "i386",
    "bootable": True,
    "checksums": {
        "sha256": "2b50438d2b96a72fac765490601d8cd24bae6ac26fa68f423b70cdb2b3bd43b3"
    },
    "disc_count": 1,
    "disc_number": 1,
    "format": "iso",
    "implant_md5": "6fae584ebfc6d462b44d2091c5425cb4",
    "mtime": 1475645667,
    "path": "Server/i386/iso/Fedora-Server-dvd-i386-25_Beta-1.1.iso",
    "size": 2108686336,
    "subvariant": "Server",
    "type": "dvd",
    "volume_id": "Fedora-S-dvd-i386-25"
}

WORKATOMIC_DICT = {
    "arch": "x86_64",
    "bootable": True,
    "checksums": {
        "sha256": "ca2aaa84009b55798eb62ab68f5d109a139f0471443e93fb600b1768555435b3"
    },
    "disc_count": 1,
    "disc_number": 1,
    "format": "iso",
    "implant_md5": "942e5efbada935b7376a63aeece53123",
    "mtime": 1475647664,
    "path": "Workstation/x86_64/iso/Fedora-Workstation-dvd-x86_64-25_Beta-1.1.iso",
    "size": 2345664512,
    "subvariant": "Workstation",
    "type": "boot",
    "volume_id": "Fedora-25-x86_64"
}

CLOUDRAWXZ_DICT = {
    "arch": "x86_64",
    "bootable": False,
    "checksums": {
        "sha256": "19e6503e29e82bea1f5a41c6a0bd773dac7a3e38507a6edc0aea3b3cf2359d18"
    },
    "disc_count": 1,
    "disc_number": 1,
    "format": "raw.xz",
    "implant_md5": "",
    "mtime": 1475647224,
    "path": "CloudImages/x86_64/images/Fedora-Cloud-Base-25_Beta-1.1.x86_64.raw.xz",
    "size": 237423412,
    "subvariant": "Cloud_Base",
    "type": "raw-xz",
    "volume_id": ""
}


@pytest.mark.usefixtures("clean_home")
class TestHelpers:
    """Tests for the functions in helpers.py."""
    def test_date_check(self):
        """Tests for date_check."""
        invalid = 'notadate'
        # this looks a bit silly, but we want the values of 'valid'
        # and 'obj' to match
        now = datetime.datetime.now()
        valid = now.strftime('%Y%m%d')
        obj = datetime.datetime.strptime(valid, '%Y%m%d')

        # Default case: checking valid obj or str should return obj
        assert fedfind.helpers.date_check(obj, out='obj') == obj
        assert fedfind.helpers.date_check(valid, out='obj') == obj
        assert fedfind.helpers.date_check(obj) == obj
        assert fedfind.helpers.date_check(valid) == obj

        # Checking valid obj or str with out='str' should return str
        assert fedfind.helpers.date_check(obj, out='str') == valid
        assert fedfind.helpers.date_check(valid, out='str') == valid

        # Checking valid with out='both' should return a tuple of both
        assert fedfind.helpers.date_check(obj, out='both') == (valid, obj)
        assert fedfind.helpers.date_check(valid, out='both') == (valid, obj)

        # Checking invalid with fail_raise=False should return False
        assert fedfind.helpers.date_check(invalid, fail_raise=False) is False

        # Checking invalid with fail_raise=True or default should
        # raise ValueError
        with pytest.raises(ValueError):
            fedfind.helpers.date_check(invalid, fail_raise=True)
        with pytest.raises(ValueError):
            fedfind.helpers.date_check(invalid)

    @mock.patch('fedfind.helpers.urlopen', return_value=True, autospec=True)
    def test_urlopen_retries_good(self, fakeopen):
        """Test urlopen_retries works and only calls urlopen once in
        normal success case.
        """
        assert fedfind.helpers.urlopen_retries(URL) is True
        fakeopen.assert_called_with(URL)
        assert fakeopen.call_count == 1

    @mock.patch('fedfind.helpers.urlopen', side_effect=ValueError, autospec=True)
    def test_urlopen_retries_bad(self, fakeopen):
        """Test urlopen_retries tries 5 times then raises ValueError
        if urlopen isn't working.
        """
        with pytest.raises(ValueError):
            fedfind.helpers.urlopen_retries(URL)
        fakeopen.assert_called_with(URL)
        assert fakeopen.call_count == 5

    @mock.patch('fedfind.helpers.urlopen', autospec=True)
    def test_download_json(self, fakeopen):
        """Test download_json decodes and returns the json provided by
        urlopen.
        """
        # this is the MagicMock that calling fakeopen returns
        fakefh = fakeopen.return_value
        # let's make it look like some JSON...
        fakefh.read.return_value = b'{"foo": "bar"}'
        assert fedfind.helpers.download_json(URL) == {'foo': 'bar'}

    @mock.patch('fedfind.helpers.urlopen', autospec=True)
    def test_url_exists_http_good(self, fakeopen):
        """Test url_exists is True when urlopen returns without error."""
        assert fedfind.helpers.url_exists(URL) is True
        fakeopen.assert_called_with(URL)

    @mock.patch('fedfind.helpers.urlopen', side_effect=ValueError('foo'), autospec=True)
    def test_url_exists_http_bad(self, fakeopen):
        """Test url_exists is False when urlopen raises error."""
        assert fedfind.helpers.url_exists(URL) is False

    def test_url_exists_invalid(self):
        """Test url_exists raises an error when the URL is garbage."""
        with pytest.raises(ValueError):
            fedfind.helpers.url_exists('jfohpjsph#^3#@^#')

    @mock.patch('fedfind.helpers.urlopen_retries', autospec=True)
    def test_get_size(self, fakeopen):
        """Test get_size works as expected, if it gets the expected
        response from urlopen_retries.
        """
        fakeresp = fakeopen.return_value
        fakeresp.info.return_value = {'Content-Length': 100}
        assert fedfind.helpers.get_size(URL) == 100

    def test_comma_list(self):
        """It splits lists on commas. It ain't rocket science."""
        assert fedfind.helpers.comma_list('foo,Bar,moo') == ['foo', 'bar', 'moo']

    def test_get_collections(self, json_collections):
        """Test _get_collections works correctly and caches."""
        (mocked, fakedata) = json_collections
        # download twice to check caching is working
        colls = fedfind.helpers._get_collections()
        colls = fedfind.helpers._get_collections()
        assert colls == fakedata
        assert mocked.call_count == 1

    def test_get_current_release(self, collections):
        """Test get_current_release gives expected values, with our
        fake collections data.
        """
        assert fedfind.helpers.get_current_release() == 23
        assert fedfind.helpers.get_current_release(branched=True) == 24

    def test_get_current_stables(self, collections):
        """Test get_current_stables gives the expected values."""
        assert fedfind.helpers.get_current_stables() == [22, 23]

    @pytest.mark.parametrize(
        ("path", "arch", "form", "typ", "subvariant", "disc_number"),
        # the image path to synthesize a dict for, and the expected values
        [
            # old-style two-week atomic compose
            ('Cloud_Atomic/x86_64/iso/Fedora-Cloud_Atomic-x86_64-23-20160313.iso',
             'x86_64', 'iso', 'dvd-ostree', 'Atomic', 1),
            ('Docker/x86_64/Fedora-Docker-Base-23-20160313.x86_64.tar.xz',
             'x86_64', 'tar.xz', 'docker', 'Docker_Base', 1),
            ('Cloud-Images/i386/Images/Fedora-Cloud-Base-23-20160313.i386.qcow2',
             'i386', 'qcow2', 'qcow2', 'Cloud_Base', 1),
            # Fedora Core 1 split discs
            ('i386/iso/yarrow-i386-disc1.iso', 'i386', 'iso', 'cd', 'Everything', 1),
            ('i386/iso/yarrow-i386-disc2.iso', 'i386', 'iso', 'cd', 'Everything', 2),
            # Various ostree cases
            # 25 Beta Workstation ostree installer
            ('Workstation/x86_64/iso/Fedora-Workstation-dvd-x86_64-25_Beta-1.1.iso',
             'x86_64', 'iso', 'dvd-ostree', 'Workstation', 1),
            # 24 two-week Atomic 20161008.0
            ('Atomic/x86_64/iso/Fedora-Atomic-dvd-x86_64-24-20161008.0.iso',
             'x86_64', 'iso', 'dvd-ostree', 'Atomic', 1),
            # F23 Final Atomic host
            ('Cloud_Atomic/x86_64/iso/Fedora-Cloud_Atomic-x86_64-23.iso',
             'x86_64', 'iso', 'dvd-ostree', 'Atomic', 1),
            # Nightly after pungi filename changes: Atomic host
            ('Atomic/x86_64/iso/Fedora-Atomic-ostree-x86_64-Rawhide-20161010.n.0.iso',
             'x86_64', 'iso', 'dvd-ostree', 'Atomic', 1),
            # Nightly after pungi filename changes: Atomic workstation
            ('Workstation/x86_64/iso/Fedora-Workstation-ostree-x86_64-Rawhide-20161010.n.0.iso',
             'x86_64', 'iso', 'dvd-ostree', 'Workstation', 1),
            # 'multi' image cases
            ('Multi/Fedora-15-Multi-Desktop.iso', '', 'iso', 'multi-desktop', 'Multi', 1),
            ('Multi/Fedora-15-Multi-Install.iso', '', 'iso', 'multi-install', 'Multi', 1),
        ]
    )
    def test_create_image_dict(self, path, arch, form, typ, subvariant, disc_number):
        """Check create_image_dict produces correct dicts with various
        standard and tricky cases.
        """
        # pylint: disable=too-many-arguments
        ret = fedfind.helpers.create_image_dict(path)
        assert ret == {
            'path': path,
            'arch': arch,
            'format': form,
            'type': typ,
            'subvariant': subvariant,
            'disc_number': disc_number
        }

    @mock.patch('fedfind.helpers.download_json', return_value=PDC_JSON, autospec=True)
    def test_pdc_query(self, fakejson):
        """Check pdc_query works correctly, with the canned PDC_JSON
        data.
        """
        assert fedfind.helpers.pdc_query('composes') == ['some', 'results']
        # first (and only) positional arg to the download_json call should be a Request
        req = fakejson.call_args[0][0]
        assert isinstance(req, Request)

    def test_find_cid(self):
        """Check find_cid works correctly, with various common and
        tricky cases.
        """
        res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora-24-20160314.n.0/compose')
        assert res == 'Fedora-24-20160314.n.0'
        res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora-24-20160314.0/compose')
        assert res == 'Fedora-24-20160314.0'
        res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora-24-20160314.t.0/compose')
        assert res == 'Fedora-24-20160314.t.0'
        res = fedfind.helpers.find_cid(
            'https://foo.com/compose/Fedora-Rawhide-20160314.n.1/compose')
        assert res == 'Fedora-Rawhide-20160314.n.1'
        res = fedfind.helpers.find_cid(
            'https://foo.com/compose/twoweek/Fedora-Atomic-24-20161004.1/compose')
        assert res == 'Fedora-Atomic-24-20161004.1'
        res = fedfind.helpers.find_cid('FedoraRespin-24-20161206.0')
        assert res == 'FedoraRespin-24-20161206.0'
        res = fedfind.helpers.find_cid('FedoraRespin-24-updates-20161206.0')
        assert res == 'FedoraRespin-24-updates-20161206.0'
        # a few 'close but no cigars' just for safety...
        res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora-24-20160314.t.u/compose')
        assert res == ''
        res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora-24-20160314/compose')
        assert res == ''
        res = fedfind.helpers.find_cid('https://foo.com/compose/Fedora/24-20160314.n.0/compose')
        assert res == ''

    @pytest.mark.parametrize(
        ("cid", "short", "version", "version_type", "base_short", "base_version", "base_type",
         "variant", "date", "compose_type", "respin"),
        [
            ('Fedora-24-20160314.n.0', 'Fedora', '24', 'ga', '', '', '', '', '20160314', 'nightly',
             0),
            ('Fedora-24-20160314.0', 'Fedora', '24', 'ga', '', '', '', '', '20160314',
             'production', 0),
            ('Fedora-24-20160314.t.0', 'Fedora', '24', 'ga', '', '', '', '', '20160314', 'test', 0),
            ('Fedora-Rawhide-20160314.n.1', 'Fedora', 'Rawhide', 'ga', '', '', '', '', '20160314',
             'nightly', 1),
            ('Fedora-Atomic-24-20160628.0', 'Fedora-Atomic', '24', 'ga', '', '', '', '',
             '20160628', 'production', 0),
            ('Fedora-20-19700101.0', 'Fedora', '20', 'ga', '', '', '', '', '19700101',
             'production', 0),
            ('FedoraRespin-24-updates-20161206.0', 'FedoraRespin', '24', 'updates', '', '', '', '',
             '20161206', 'production', 0),
            # older fake Respin CID
            ('FedoraRespin-24-20161206.0', 'FedoraRespin', '24', 'ga', '', '', '', '', '20161206',
             'production', 0),
            # taken from productmd test_composeinfo
            ('Fedora-22-20160622.0', 'Fedora', '22', 'ga', '', '', '', '', '20160622', 'production',
             0),
            # FIXME: 'ci' is a valid compose type now but get_date_type_respin doesn't handle it
            # 'Fedora-22-20160622.ci.0'
            ('Fedora-22-updates-20160622.0', 'Fedora', '22', 'updates', '', '', '', '', '20160622',
             'production', 0),
            ('Fedora-22-updates-20160622.n.0', 'Fedora', '22', 'updates', '', '', '', '',
             '20160622', 'nightly', 0),
            ('Fedora-22-BASE-3-20160622.0', 'Fedora', '22', 'ga', 'BASE', '3', 'ga', '', '20160622',
             'production', 0),
            ('Fedora-22-BASE-3-20160622.n.0', 'Fedora', '22', 'ga', 'BASE', '3', 'ga', '',
             '20160622', 'nightly', 0),
            ('Fedora-22-updates-BASE-3-20160622.0', 'Fedora', '22', 'updates', 'BASE', '3', 'ga',
             '', '20160622', 'production', 0),
            ('Fedora-22-updates-BASE-3-20160622.n.0', 'Fedora', '22', 'updates', 'BASE', '3', 'ga',
             '', '20160622', 'nightly', 0),
            ('Fedora-22-BASE-3-updates-20160622.0', 'Fedora', '22', 'ga', 'BASE', '3', 'updates',
             '', '20160622', 'production', 0),
            ('Fedora-22-BASE-3-updates-20160622.n.0', 'Fedora', '22', 'ga', 'BASE', '3', 'updates',
             '', '20160622', 'nightly', 0),
            ('Fedora-22-updates-BASE-3-updates-20160622.0', 'Fedora', '22', 'updates', 'BASE', '3',
             'updates', '', '20160622', 'production', 0),
            ('Fedora-22-updates-BASE-3-updates-20160622.n.0', 'Fedora', '22', 'updates', 'BASE',
             '3', 'updates', '', '20160622', 'nightly', 0),
            ('Fedora-22-updates-BASE-3-updates-testing-20160622.n.0', 'Fedora', '22',
             'updates', 'BASE', '3', 'updates-testing', '', '20160622', 'nightly', 0),
            ('Fedora-22-updates-testing-BASE-3-updates-20160622.n.0', 'Fedora', '22',
             'updates-testing', 'BASE', '3', 'updates', '', '20160622', 'nightly', 0),
            ('Fedora-22-updates-testing-BASE-3-updates-testing-20160622.n.0', 'Fedora', '22',
             'updates-testing', 'BASE', '3', 'updates-testing', '', '20160622', 'nightly', 0),
            # Rawhide complex
            ('Fedora-Rawhide-updates-RHEL-6.3.4-20160513.t.1', 'Fedora', 'Rawhide', 'updates',
             'RHEL', '6.3.4', 'ga', '', '20160513', 'test', 1),
            # I think this is how a RHEL 5 one looks?
            ('Fedora-5-updates-Server-20160523.2', 'Fedora', '5', 'updates', '', '', '', 'Server',
             '20160523', 'production', 2),
            ('Fedora-5-updates-Client-20160523.2', 'Fedora', '5', 'updates', '', '', '', 'Client',
             '20160523', 'production', 2),
        ]
    )
    def test_parse_cid(self, cid, short, version, version_type, base_short, base_version,
                       base_type, variant, date, compose_type, respin):
        """Check parse_cid works correctly, with various known compose
        ID forms.
        """
        # pylint: disable=too-many-arguments
        assert fedfind.helpers.parse_cid(cid, dic=True) == {
            'short': short,
            'version': version,
            'version_type': version_type,
            'base_short': base_short,
            'base_version': base_version,
            'base_type': base_type,
            'variant': variant,
            'date': date,
            'compose_type': compose_type,
            'respin': respin
        }

    def test_parse_cid_tuple(self):
        """Check the two older tuple formats for parse_cid work as
        expected.
        """
        res = fedfind.helpers.parse_cid('Fedora-24-20160314.n.0')
        assert res == ('24', '20160314', 'nightly', 0)
        res = fedfind.helpers.parse_cid('Fedora-24-20160314.n.0', dist=True)
        assert res == ('Fedora', '24', '20160314', 'nightly', 0)
        res = fedfind.helpers.parse_cid('Fedora-Atomic-24-20160628.0', dist=True)
        assert res == ('Fedora-Atomic', '24', '20160628', 'production', 0)

    def test_parse_cid_error(self):
        """Check parse_cid raises ValueError for non-Pungi 4-ish CID."""
        with pytest.raises(ValueError):
            fedfind.helpers.parse_cid('23-20160530')

    @mock.patch('fedfind.helpers.download_json', return_value=PDC_JSON_REAL, autospec=True)
    def test_cid_from_label(self, fakejson):
        """Check cid_from_label works, using some real canned PDC
        data.
        """
        assert fedfind.helpers.cid_from_label('24', 'Alpha-1.1') == 'Fedora-24-20160314.1'

    @mock.patch('fedfind.helpers.download_json', return_value=PDC_JSON_REAL, autospec=True)
    def test_label_from_cid(self, fakejson):
        """Check label_from_cid works, using some real canned PDC
        data.
        """
        assert fedfind.helpers.label_from_cid('Fedora-24-20160314.1') == 'Alpha-1.1'

    def test_correct_image(self):
        """Test the correct_image function does all the corrections
        it ought to.
        """
        # dict with no 'issues' should generate identical result
        assert fedfind.helpers.correct_image(SERVDVD_DICT) == SERVDVD_DICT
        # Atomic installer dict should be modified as expected
        newdic = fedfind.helpers.correct_image(WORKATOMIC_DICT)
        # type should be changed
        assert newdic['type'] == 'dvd-ostree'
        # otherwise dict should be the same
        newdic['type'] = 'boot'
        assert newdic == WORKATOMIC_DICT

    def test_identify_image(self):
        """Test that identify_image provides the correct tuples with
        various image dicts, and respects its options correctly.
        """
        ident = fedfind.helpers.identify_image
        assert ident(SERVDVD_DICT) == ('Server', 'dvd', 'iso')
        assert ident(WORKATOMIC_DICT) == ('Workstation', 'dvd-ostree', 'iso')
        # test options
        assert ident(SERVDVD_DICT, out='tuple') == ('Server', 'dvd', 'iso')
        assert ident(SERVDVD_DICT, out='string') == ('Server-dvd-iso')
        assert ident(SERVDVD_DICT, lower=True) == ('server', 'dvd', 'iso')
        assert ident(CLOUDRAWXZ_DICT, undersub=False) == ('Cloud_Base', 'raw-xz', 'raw.xz')
        assert ident(CLOUDRAWXZ_DICT, undersub=True) == ('Cloud_Base', 'raw_xz', 'raw.xz')

    @pytest.mark.parametrize("flist", ('fedora', 'alt', 'archive'))
    def test_get_filelist(self, flist, http, clean_home):
        """Test _get_filelist with each of the filelists, in various
        ways (caching, failed downloads, etc.)
        """
        def check_file(flist, touch=False):
            """Inner function for doing the check (run multiple times)"""
            origfile = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'http',
                                    'pub', flist, 'imagelist-{0}'.format(flist))
            if touch:
                os.utime(origfile, None)
            gotcksum = hashlib.md5()
            origcksum = hashlib.md5()
            record = {}
            with fedfind.helpers._get_filelist(flist=flist, record=record) as testfh:
                assert testfh.name == os.path.join(clean_home, '.cache', 'fedfind',
                                                   'imagelist-{0}'.format(flist))
                for line in testfh:
                    gotcksum.update(line.encode('utf-8'))
            with codecs.open(origfile, encoding='utf-8') as origfh:
                for line in origfh:
                    origcksum.update(line.encode('utf-8'))
            assert gotcksum.digest() == origcksum.digest()
            return record

        check_file(flist)
        record = check_file(flist)
        assert record['cache'] is True

        record = check_file(flist, touch=True)
        assert record['cache'] is False

        # now make the download fail: we should still work using the
        # cached copy, but warn about it
        origconst = fedfind.const.HTTPS_DL
        fedfind.const.HTTPS_DL = 'http://localhost:5001/nonexistent'
        check_file(flist)
        # Reset the constant
        fedfind.const.HTTPS_DL = origconst

    @mock.patch('fedfind.helpers.download_json', return_value=PDC_JSON, autospec=True)
    def test_cachedir_vanish(self, http, clean_home):
        """Test that we recover from the cache dir suddenly vanishing.
        I saw this happen on openqa01 once - I think when we're using
        a temp dir as the cache location, it can get cleaned up by
        systemd or so while we're running.
        """
        fedfind.helpers._get_filelist(flist='fedora')
        assert os.path.exists(os.path.join(clean_home, '.cache', 'fedfind'))
        shutil.rmtree(os.path.join(clean_home, '.cache', 'fedfind'))
        fedfind.helpers._get_filelist(flist='fedora')
        assert os.path.exists(os.path.join(clean_home, '.cache', 'fedfind'))
        # also test pdc_query
        shutil.rmtree(os.path.join(clean_home, '.cache', 'fedfind'))
        # pass compose_id otherwise the function won't cache
        fedfind.helpers.pdc_query('composes', {'compose_id': 'Fedora-26-20170418.n.0'})
        assert os.path.exists(os.path.join(clean_home, '.cache', 'fedfind'))

# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2016 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=old-style-class, no-init, protected-access, no-self-use, unused-argument

"""Test configuration and fixtures."""

from __future__ import unicode_literals
from __future__ import print_function

import json
import os
import re
import shutil
import subprocess
import time

import mock
import pytest

import fedfind.const
import fedfind.helpers

# Fake data for the collections fixture
COLLECTIONS_JSON = [
    {
        "allow_retire": True,
        "branchname": "master",
        "date_created": "2014-05-14 12:36:15",
        "date_updated": "2016-02-23 22:56:34",
        "dist_tag": ".fc25",
        "koji_name": "rawhide",
        "name": "Fedora",
        "status": "Under Development",
        "version": "devel"
    },
    {
        "allow_retire": False,
        "branchname": "f22",
        "date_created": "2015-02-10 14:00:01",
        "date_updated": "2015-02-10 14:00:01",
        "dist_tag": ".fc22",
        "koji_name": "f22",
        "name": "Fedora",
        "status": "Active",
        "version": "22"
    },
    {
        "allow_retire": False,
        "branchname": "f23",
        "date_created": "2015-07-14 18:13:12",
        "date_updated": "2015-07-14 18:13:12",
        "dist_tag": ".fc23",
        "koji_name": "f23",
        "name": "Fedora",
        "status": "Active",
        "version": "23"
    },
    {
        "allow_retire": True,
        "branchname": "f24",
        "date_created": "2016-02-23 22:57:55",
        "date_updated": "2016-02-25 20:39:53",
        "dist_tag": ".fc24",
        "koji_name": "f24",
        "name": "Fedora",
        "status": "Under Development",
        "version": "24"
    }
]

@pytest.fixture(scope="session")
def http(request):
    """Run a SimpleHTTPServer that we can use as a fake dl.fp.o. Serve
    the contents of tests/data/http, for the entire test session. We
    just do this with subprocess as we need it to run parallel to the
    tests and this is really the easiest way.
    """
    root = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        'data', 'http')
    # Always use python2 for this (for now)
    args = ("python2", "-m", "SimpleHTTPServer", "5001")
    proc = subprocess.Popen(args, cwd=root)
    # this is rather ugly, but if we yield immediately, the tests will
    # try to use the server before it's actually started up. I don't
    # know of a way to wait for the sub-process to print a specific
    # message but *not* wait for it to complete (communicate() waits
    # for completion), so just wait a second.
    time.sleep(1)
    # Redefine the HTTPS_DL and HTTPS constants to point to the fake
    fedfind.const.HTTPS_DL = 'http://localhost:5001/pub'
    fedfind.const.HTTPS = 'http://localhost:5001/pub'

    def fin():
        """Teardown: kill the server. We don't use the 'yield fixture'
        feature here because we want these tests to run on EL 6, and
        the steam-powered ancient pytest in EL 6 doesn't have it.
        """
        proc.kill()
    request.addfinalizer(fin)

    return None

@pytest.fixture(scope="function")
def clean_home(request):
    """Provides a fake user home directory, at data/home/ under the
    tests directory. Before the test, re-create it, change the cache
    dir global to use it, and patch os.path.expanduser to return it.
    After the test, delete it and clean up the other bits. Yields
    the full path.
    """
    home = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'home')
    if os.path.exists(home):
        shutil.rmtree(home)
    cachedir = os.path.join(home, '.cache', 'fedfind')
    os.makedirs(cachedir)
    patcher = mock.patch('os.path.expanduser', return_value=home, autospec=True)
    patcher.start()
    origcache = fedfind.helpers.CACHEDIR
    fedfind.helpers.CACHEDIR = cachedir

    def fin():
        """Teardown: stop the patcher and wipe the directory. As per
        http, we can't use the yield syntax here because EL 6.
        """
        patcher.stop()
        fedfind.helpers.CACHEDIR = origcache
        if os.path.exists(home):
            shutil.rmtree(home)
    request.addfinalizer(fin)

    return home

def fake_pdc_query(path, params=None):
    """A fake pdc_query that doesn't really hit PDC, it just returns
    canned information (some inline, some loaded from tests/data) for
    queries the tests are known to make.
    """
    # match how pdc_query sanitizes the path
    path = path.strip('/') + '/'
    # composes query. This is hit for cid_from_label.
    if path == 'composes/':
        # Knowns in (release, label, composeid) form:
        composes = (
            ('fedora-24', 'RC-1.2', 'Fedora-24-20160614.0'),
            ('fedora-25', 'RC-1.3', 'Fedora-25-20161115.0'),
            ('fedora-26', 'RC-1.5', 'Fedora-26-20170705.0'),
            ('fedora-27', 'RC-1.6', 'Fedora-27-20171105.0'),
        )
        for (rel, label, cid) in composes:
            # We can make this matching smarter if needed, for now
            # it's specific to cid_from_label
            if params['release'] == rel and params['compose_label'] == label:
                return [{'compose_id': cid, 'release': rel, 'compose_label': label}]
        # fallthrough
        return []

    # compose-images query. Just grab the compose ID and look for a
    # backing file to return.
    compimgs = re.compile(r'compose-images/(.*)/')
    match = compimgs.match(path)
    if match:
        fname = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             'data', 'pdc-compimages-{0}.json'.format(match.group(1)))
        with open(fname, 'r') as datafh:
            return json.loads(datafh.read())

@pytest.fixture(scope="function")
def pdc(request):
    """Mock fedfind.helpers.pdc_query with fake_pdc_query."""
    patcher = mock.patch('fedfind.helpers.pdc_query', fake_pdc_query)
    patcher.start()

    def fin():
        """Teardown: stop the patcher and wipe the directory. As per
        http, we can't use the yield syntax here because EL 6.
        """
        patcher.stop()
    request.addfinalizer(fin)

    return None

@pytest.fixture(scope="function")
def collections(request):
    """Mock fedfind.helpers._get_collections to return test data."""
    patcher = mock.patch('fedfind.helpers._get_collections', return_value=COLLECTIONS_JSON,
                         autospec=True)
    patcher.start()

    def fin():
        """Teardown: stop the patcher and wipe the directory. As per
        http, we can't use the yield syntax here because EL 6.
        """
        patcher.stop()
    request.addfinalizer(fin)

    return None

@pytest.fixture(scope="function")
def json_collections(request):
    """Mock fedfind.helpers.download_json to return collections test
    data (wrapped appropriately). Returns the mock so tests can check
    its call count, and the test data so they can verify it.
    """
    patcher = mock.patch(
        'fedfind.helpers.download_json', return_value={'collections': COLLECTIONS_JSON},
        autospec=True)
    mocked = patcher.start()

    def fin():
        """Teardown: stop the patcher and wipe the directory. As per
        http, we can't use the yield syntax here because EL 6.
        """
        patcher.stop()
    request.addfinalizer(fin)

    return (mocked, COLLECTIONS_JSON)

# vim: set textwidth=100 ts=8 et sw=4:



#!/bin/python3

"""This script generates the compose-urls-YYYYMMDD.json list of compose
URLs that is used for the test_compose_urls test. You need to have a
checkout of fedora_nightlies symlinked to wherever you run the script
from for it to work.
"""

import datetime
import logging
import fedora_nightlies
import json

logging.basicConfig(level=logging.DEBUG)

msgs = fedora_nightlies.datagrepper_query(['org.fedoraproject.prod.pungi.compose.status.change'], 180)
urls = [msg.get('location') for msg in msgs]
urls = set([url for url in urls if url])

fn = 'compose-urls-{0}.json'.format(datetime.date.today().strftime('%Y%m%d'))
fh = open(fn, 'w')
json.dump(list(urls), fh, indent=4, separators=(',', ': '))
fh.close()



# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Defines custom exceptions used by fedfind."""

from __future__ import unicode_literals
from __future__ import print_function

class FedfindError(Exception):
    """Base class for all fedfind errors."""
    pass


class PreviousRC1Error(FedfindError):
    """Special exception for trying to get previous_release of a TC1
    or RC1. Makes it easy to spot this case."""
    pass


class WaitError(FedfindError):
    """Error raised when some kind of wait has gone on too long."""
    pass


class DiscoveryError(FedfindError):
    """Raised when some kind of discovery attempt failed."""
    pass


class UnsupportedComposeError(FedfindError):
    """Raised when get_release decides this is a compose it knows we
    don't support.
    """
    pass


class UrlMatchError(FedfindError):
    """Raised when get_release(url=someurl) gets a Release instance,
    but its location does not match someurl.
    """
    def __init__(self, url, release):
        self.url = url
        self.release = release
        super(UrlMatchError, self).__init__()

    def __str__(self):
        tmpl = "URL mismatch! URL: {0} Class: {1} Location: {2}"
        return tmpl.format(self.url, self.release.__class__.__name__, self.release.location)


class CidMatchError(FedfindError):
    """Raised when get_release(cid=somecid) gets a Release instance,
    but its compose ID does not match somecid.
    """
    def __init__(self, cid, release):
        self.cid = cid
        self.release = release
        super(CidMatchError, self).__init__()

    def __str__(self):
        tmpl = "Compose ID mismatch! Requested CID: {0} Class: {1} Class CID: {2}"
        return tmpl.format(self.cid, self.release.__class__.__name__, self.release.cid)

# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Various shared constants."""

from __future__ import unicode_literals
from __future__ import print_function

import collections
import re

# For all of these, we identify each we want to consider unique; that
# is the 'name'. 'values' is an iterable of values which should map to
# that name; for instance, we don't want to treat i486, i586 and i686
# as separate arches, but every time we see any of those we want to
# treat it as i386. 'weight' is an integer to be used for sort
# weighting if desired.
#
# For ArchTuple, each entry in ARCHES is a primary arch. group is used
# to group related arches together. current is whether it's currently
# a primary arch.
#
# For LoadTuple, each entry in FLAVORS or LOADOUTS is something we
# consider a unique flavor or loadout. subs is an iterable of sub-
# flavors.

ArchTuple = collections.namedtuple('ArchTuple',
                                   'name values group current')
LoadTuple = collections.namedtuple('LoadTuple', 'name values')


HTTPS = 'https://download.fedoraproject.org/pub'
HTTPS_DL = 'https://dl.fedoraproject.org/pub'
NIGHTLY_BASE = 'https://kojipkgs.fedoraproject.org/compose'
KNOWN_PREFIXES = (HTTPS, 'http://download.fedoraproject.org/pub',
                  HTTPS_DL, 'http://dl.fedoraproject.org/pub',
                  NIGHTLY_BASE, 'http://kojipkgs.fedoraproject.org/compose',
                  # secondary arch compose locations, see RHBZ #1435953
                  'http://ppc.koji.fedoraproject.org/compose',
                  'https://ppc.koji.fedoraproject.org/compose',
                  'http://s390.koji.fedoraproject.org/compose',
                  'https://s390.koji.fedoraproject.org/compose',
                  'http://arm.koji.fedoraproject.org/compose',
                  'https://arm.koji.fedoraproject.org/compose',)
PDC_API = 'https://pdc.fedoraproject.org/rest_api/v1'

# we could import these from pungi but jeez, the dependencies are
# growing as it is. Pungi compose states.
PUNGI_SUCCESS = ("FINISHED", "FINISHED_INCOMPLETE")
PUNGI_FAILURE = ("DOOMED",)
PUNGI_DONE = PUNGI_SUCCESS + PUNGI_FAILURE

# SUBVARIANTS
# used for synthesizing 'subvariant' value for pre-Pungi 4 composes
# This has to be ordered so we get the right result for e.g. 'kde'
# vs. 'jam_kde'
SUBVARIANTS = (
    LoadTuple('Desktop', ('desktop',)),
    LoadTuple('Multi', ('multi',)),
    LoadTuple('KDE', ('kde',)),
    LoadTuple('Xfce', ('xfce',)),
    LoadTuple('SoaS', ('soas',)),
    LoadTuple('Mate', ('mate_compiz', 'mate-compiz', 'mate')),
    LoadTuple('Cinnamon', ('cinnamon', 'cinn')),
    LoadTuple('LXDE', ('lxde',)),
    LoadTuple('Design_suite', ('design_suite', 'design-suite')),
    LoadTuple('Electronic_lab', ('electronic_lab', 'electronic-lab')),
    LoadTuple('Games', ('games',)),
    LoadTuple('Robotics', ('robotics',)),
    LoadTuple('Security', ('security',)),
    LoadTuple('Jam_KDE', ('jam_kde',)),
    LoadTuple('Scientific_KDE', ('scientific_kde', 'scientific-kde')),
    LoadTuple('Astronomy_KDE', ('astronomy_kde',)),
    LoadTuple('Minimal', ('minimal',)),
    LoadTuple('Source', ('source', 'srpms')),
    LoadTuple('Cloud', ('cloud', 'docker-base')),
    LoadTuple('Cloud_Base', ('cloud-base', 'cloud_base')),
    # the '-' here is to prevent Postrelease nightly Cloud images
    # getting 'Atomic' as their subvariant, as they always have
    # /atomic/ in their path names
    LoadTuple('Atomic', ('atomic-',)),
    LoadTuple('Container_Minimal_Base', ('container-minimal-base',)),
    LoadTuple('Docker_Base', ('docker-base',)),
    LoadTuple('Workstation', ('workstation', 'work')),
    LoadTuple('Workstation Ostree', ('workstationostree',)),
    LoadTuple('Server', ('server',)),
    LoadTuple('Everything', ('everything',)),
    LoadTuple('Python_Classroom', ('python-classroom',)),
    LoadTuple('LXQt', ('lxqt',)),
)

# ARCHES
ARCHES = (
    ArchTuple('x86_64', ('x86_64',), 'intel', True),
    ArchTuple('i386', ('i386', 'i486', 'i586', 'i686'), 'intel', False),
    ArchTuple('armhfp', ('armhfp',), 'arm', True),
    ArchTuple('ppc', ('ppc',), 'ppc', False),
    ArchTuple('ppc64', ('ppc64',), 'ppc', False)
)

# regex used to identify live-respins images. Used in a few different
# places, so we keep it here. Uses named groups for convenience

RESPINRE = re.compile(r'F(?P<release>\d+)-(?P<subv>[^-]+)-(?P<arch>[^-]+)-(?P<date>\d{8,8})\.iso',
                      re.I)

# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=old-style-class, no-init, protected-access, no-self-use, unused-argument
# pylint: disable=invalid-name, too-few-public-methods, too-many-public-methods, too-many-lines

"""Tests for release.py."""

from __future__ import unicode_literals
from __future__ import print_function

import codecs
import datetime
import json
import os
import pytest

import mock
import six

import fedfind.release

# Some chunks of HTML from mirror package directories for testing
# get_package_nvras. FC6 has generic release (no fcXX), F16 has
# fcXX, F21 is split by initials. FC6 is from a mirror, F18 and F21
# are from dl.fp.o.

FC6PACKAGEHTML = """
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
<head>
<title>Index of /fedora/archive/fedora/linux/core/6/source/SRPMS/</title>
<style type="text/css">
a, a:active {text-decoration: none; color: blue;}
a:visited {color: #48468F;}
a:hover, a:focus {text-decoration: underline; color: red;}
body {background-color: #F5F5F5;}
h2 {margin-bottom: 12px;}
table {margin-left: 12px;}
th, td { font: 90% monospace; text-align: left;}
th { font-weight: bold; padding-right: 14px; padding-bottom: 3px;}
td {padding-right: 14px;}
td.s, th.s {text-align: right;}
div.list { background-color: white; border-top: 1px solid #646464; border-bottom: 1px solid #646464; padding-top: 10px; padding-bottom: 14px;}
div.foot { font: 90% monospace; color: #787878; padding-top: 4px;}
</style>
</head>
<body>
<h2>Index of /fedora/archive/fedora/linux/core/6/source/SRPMS/</h2>
<div class="list">
<table summary="Directory Listing" cellpadding="0" cellspacing="0">
<thead><tr><th class="n">Name</th><th class="m">Last Modified</th><th class="s">Size</th><th class="t">Type</th></tr></thead>
<tbody>
<tr><td class="n"><a href="../">Parent Directory</a>/</td><td class="m">&nbsp;</td><td class="s">- &nbsp;</td><td class="t">Directory</td></tr>
<tr><td class="n"><a href="repodata/">repodata</a>/</td><td class="m">2006-Oct-17 21:22:24</td><td class="s">- &nbsp;</td><td class="t">Directory</td></tr>
<tr><td class="n"><a href="am-utils-6.1.5-4.src.rpm">am-utils-6.1.5-4.src.rpm</a></td><td class="m">2006-Oct-05 17:06:36</td><td class="s">1.8M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="amanda-2.5.0p2-4.src.rpm">amanda-2.5.0p2-4.src.rpm</a></td><td class="m">2006-Oct-05 14:53:24</td><td class="s">1.7M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="amtu-1.0.4-3.1.src.rpm">amtu-1.0.4-3.1.src.rpm</a></td><td class="m">2006-Oct-05 15:01:09</td><td class="s">58.9K</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="anaconda-11.1.1.3-1.src.rpm">anaconda-11.1.1.3-1.src.rpm</a></td><td class="m">2006-Oct-17 19:46:31</td><td class="s">3.7M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="anacron-2.3-41.fc6.src.rpm">anacron-2.3-41.fc6.src.rpm</a></td><td class="m">2006-Oct-12 23:23:17</td><td class="s">42.0K</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="ant-1.6.5-2jpp.2.src.rpm">ant-1.6.5-2jpp.2.src.rpm</a></td><td class="m">2006-Oct-05 16:00:10</td><td class="s">5.0M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="anthy-7900-2.fc6.src.rpm">anthy-7900-2.fc6.src.rpm</a></td><td class="m">2006-Oct-05 17:13:24</td><td class="s">4.4M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="antlr-2.7.6-4jpp.2.src.rpm">antlr-2.7.6-4jpp.2.src.rpm</a></td><td class="m">2006-Oct-05 14:54:40</td><td class="s">1.3M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="apmd-3.2.2-5.src.rpm">apmd-3.2.2-5.src.rpm</a></td><td class="m">2006-Oct-05 17:06:38</td><td class="s">99.6K</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="apr-1.2.7-10.src.rpm">apr-1.2.7-10.src.rpm</a></td><td class="m">2006-Oct-05 17:04:36</td><td class="s">1.0M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="basesystem-8.0-5.1.1.src.rpm">basesystem-8.0-5.1.1.src.rpm</a></td><td class="m">2006-Oct-05 14:53:05</td><td class="s">3.7K</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="bash-3.1-16.1.src.rpm">bash-3.1-16.1.src.rpm</a></td><td class="m">2006-Oct-05 17:08:36</td><td class="s">4.4M</td><td class="t">application/x-rpm</td></tr>
<tr><td class="n"><a href="bc-1.06-21.src.rpm">bc-1.06-21.src.rpm</a></td><td class="m">2006-Oct-05 18:46:45</td><td class="s">236.9K</td><td class="t">application/x-rpm</td></tr>
"""

F16PACKAGEHTML = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<html>
 <head>
  <title>Index of /pub/archive/fedora/linux/releases/16/Everything/source/SRPMS</title>
 </head>
 <body>
<h1>Index of /pub/archive/fedora/linux/releases/16/Everything/source/SRPMS</h1>
<pre><img src="/icons/blank.gif" alt="Icon "> <a href="?C=N;O=D">Name</a>                                                                         <a href="?C=M;O=A">Last modified</a>      <a href="?C=S;O=A">Size</a>  <a href="?C=D;O=A">Description</a><hr><img src="/icons/back.gif" alt="[PARENTDIR]"> <a href="/pub/archive/fedora/linux/releases/16/Everything/source/">Parent Directory</a>                                                                                  -   
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amanda-3.3.0-2.fc16.src.rpm">amanda-3.3.0-2.fc16.src.rpm</a>                                                  2011-07-30 03:39  4.0M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amanith-0.3-17.fc16.src.rpm">amanith-0.3-17.fc16.src.rpm</a>                                                  2011-07-30 03:18  7.3M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amarok-2.4.3-1.fc16.src.rpm">amarok-2.4.3-1.fc16.src.rpm</a>                                                  2011-08-02 02:53   17M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amavisd-new-2.6.6-1.fc16.src.rpm">amavisd-new-2.6.6-1.fc16.src.rpm</a>                                             2011-09-19 17:12  950K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amide-1.0.0-1.fc16.src.rpm">amide-1.0.0-1.fc16.src.rpm</a>                                                   2011-10-11 00:33  1.5M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amoebax-0.2.0-7.fc15.src.rpm">amoebax-0.2.0-7.fc15.src.rpm</a>                                                 2011-07-30 01:29   10M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amora-1.1-6.fc15.src.rpm">amora-1.1-6.fc15.src.rpm</a>                                                     2011-07-30 02:16  161K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amqp-1.0.819819-2.fc15.src.rpm">amqp-1.0.819819-2.fc15.src.rpm</a>                                               2011-07-30 03:46  225K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amsn-0.98.4-4.fc16.src.rpm">amsn-0.98.4-4.fc16.src.rpm</a>                                                   2011-07-30 04:03   13M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amtterm-1.3-1.fc16.src.rpm">amtterm-1.3-1.fc16.src.rpm</a>                                                   2011-07-30 03:51   43K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amtu-1.0.8-8.fc15.src.rpm">amtu-1.0.8-8.fc15.src.rpm</a>                                                    2011-07-30 05:09  142K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="anaconda-16.25-1.fc16.src.rpm">anaconda-16.25-1.fc16.src.rpm</a>                                                2011-11-03 02:14  5.1M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="anaconda-yum-plugins-1.0-6.fc15.src.rpm">anaconda-yum-plugins-1.0-6.fc15.src.rpm</a>                                      2011-07-30 02:19   14K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="basesystem-10.0-5.fc16.src.rpm">basesystem-10.0-5.fc16.src.rpm</a>                                               2011-07-30 02:47  5.7K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="bash-4.2.10-4.fc16.src.rpm">bash-4.2.10-4.fc16.src.rpm</a>                                                   2011-07-30 04:23  6.7M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="bash-completion-1.3-6.fc16.src.rpm">bash-completion-1.3-6.fc16.src.rpm</a>                                           2011-09-06 16:43  233K  
<hr></pre>
"""

F18PACKAGEHTMLA = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<html>
 <head>
  <title>Index of /pub/archive/fedora/linux/releases/18/Everything/source/SRPMS/a</title>
 </head>
 <body>
<h1>Index of /pub/archive/fedora/linux/releases/18/Everything/source/SRPMS/a</h1>
<pre><img src="/icons/blank.gif" alt="Icon "> <a href="?C=N;O=D">Name</a>                                                           <a href="?C=M;O=A">Last modified</a>      <a href="?C=S;O=A">Size</a>  <a href="?C=D;O=A">Description</a><hr><img src="/icons/back.gif" alt="[PARENTDIR]"> <a href="/pub/archive/fedora/linux/releases/18/Everything/source/SRPMS/">Parent Directory</a>                                                                    -   
<img src="/icons/unknown.gif" alt="[   ]"> <a href="am-utils-6.1.5-23.fc18.src.rpm">am-utils-6.1.5-23.fc18.src.rpm</a>                                 2012-08-11 04:16  1.9M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amanda-3.3.2-2.fc18.src.rpm">amanda-3.3.2-2.fc18.src.rpm</a>                                    2012-09-18 17:58  4.2M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amanith-0.3-22.fc18.src.rpm">amanith-0.3-22.fc18.src.rpm</a>                                    2012-08-11 04:50  7.3M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amarok-2.6.0-4.fc18.src.rpm">amarok-2.6.0-4.fc18.src.rpm</a>                                    2012-09-12 18:15   40M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amavisd-new-2.8.0-2.fc18.src.rpm">amavisd-new-2.8.0-2.fc18.src.rpm</a>                               2012-08-11 06:33  1.0M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="ambdec-0.5.1-3.fc18.src.rpm">ambdec-0.5.1-3.fc18.src.rpm</a>                                    2012-08-11 06:47  225K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amide-1.0.0-3.fc18.src.rpm">amide-1.0.0-3.fc18.src.rpm</a>                                     2012-08-11 04:36  1.5M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amoebax-0.2.0-10.fc18.src.rpm">amoebax-0.2.0-10.fc18.src.rpm</a>                                  2012-08-11 04:41   10M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amora-1.1-9.fc18.src.rpm">amora-1.1-9.fc18.src.rpm</a>                                       2012-11-02 03:48  161K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amqp-1.0.819819-4.fc18.src.rpm">amqp-1.0.819819-4.fc18.src.rpm</a>                                 2012-08-11 07:05  225K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="ams-2.0.1-5.fc18.src.rpm">ams-2.0.1-5.fc18.src.rpm</a>                                       2012-10-26 06:52  291K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amsn-0.98.9-4.fc18.src.rpm">amsn-0.98.9-4.fc18.src.rpm</a>                                     2012-08-11 07:22   13M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amtterm-1.3-4.fc18.src.rpm">amtterm-1.3-4.fc18.src.rpm</a>                                     2012-08-11 05:06   44K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="amtu-1.0.8-12.fc18.src.rpm">amtu-1.0.8-12.fc18.src.rpm</a>                                     2012-08-11 04:53  143K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="anaconda-18.37.11-1.fc18.src.rpm">anaconda-18.37.11-1.fc18.src.rpm</a>                               2013-01-08 04:12  3.7M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="anaconda-yum-plugins-1.0-8.fc18.src.rpm">anaconda-yum-plugins-1.0-8.fc18.src.rpm</a>                        2012-08-11 04:20   14K  
"""

F18PACKAGEHTMLB = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<html>
 <head>
  <title>Index of /pub/archive/fedora/linux/releases/18/Everything/source/SRPMS/b</title>
 </head>
 <body>
<h1>Index of /pub/archive/fedora/linux/releases/18/Everything/source/SRPMS/b</h1>
<pre><img src="/icons/blank.gif" alt="Icon "> <a href="?C=N;O=D">Name</a>                                                           <a href="?C=M;O=A">Last modified</a>      <a href="?C=S;O=A">Size</a>  <a href="?C=D;O=A">Description</a><hr><img src="/icons/back.gif" alt="[PARENTDIR]"> <a href="/pub/archive/fedora/linux/releases/18/Everything/source/SRPMS/">Parent Directory</a>                                                                    -   
<img src="/icons/unknown.gif" alt="[   ]"> <a href="basesystem-10.0-7.fc18.src.rpm">basesystem-10.0-7.fc18.src.rpm</a>                                 2012-08-11 05:28  6.0K  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="bash-4.2.39-3.fc18.src.rpm">bash-4.2.39-3.fc18.src.rpm</a>                                     2012-11-30 03:56  6.8M  
<img src="/icons/unknown.gif" alt="[   ]"> <a href="bash-completion-2.0-2.fc18.src.rpm">bash-completion-2.0-2.fc18.src.rpm</a>
"""

# Fake PDC response for get_package_nvras testing.

PDCPACKAGES = [
    {
        "id": 222481,
        "name": "anaconda",
        "version": "24.13",
        "epoch": 0,
        "release": "1.fc24",
        "arch": "src",
        "srpm_name": "anaconda",
        "srpm_nevra": "anaconda-0:24.13-1.fc24.src",
        "filename": "anaconda-24.13-1.fc24.src.rpm",
    },
    {
        "id": 120399,
        "name": "bash",
        "version": "4.3.42",
        "epoch": 0,
        "release": "4.fc24",
        "arch": "src",
        "srpm_name": "bash",
        "srpm_nevra": "bash-0:4.3.42-4.fc24.src",
        "filename": "bash-4.3.42-4.fc24.src.rpm",
    },
    {
        "id": 222118,
        "name": "amanda",
        "version": "3.3.8",
        "epoch": 0,
        "release": "1.fc24",
        "arch": "src",
        "srpm_name": "amanda",
        "srpm_nevra": "amanda-0:3.3.8-1.fc24.src",
        "filename": "amanda-3.3.8-1.fc24.src.rpm",
    }
]


class FakeRelease(object):
    """Release stub used for testing nightly respin guessing. Gives
    the magic 'good' values for exists and status if respin is 3 and
    release is 24 or Rawhide.
    """
    def __init__(self, respin='', *args, **kwargs):
        self.exists = False
        self.status = 'DOOMED'
        self.respin = respin
        if respin == 3:
            self.exists = True
            self.status = 'FINISHED'


class FakeResponsePackages(object):
    """urlopen response stub used for get_package_nvr testing."""
    def __init__(self, url):
        self.url = url

    def read(self):
        """Returns the appropriate chunk of HTML (see above) if the
        correct URL was given.
        """
        arcbase = '{0}/archive/fedora/linux/'.format(fedfind.const.HTTPS)
        branchbase = 'https://kojipkgs.fedoraproject.org/compose/branched/Fedora-24-20160321.n.0/'
        modbase = 'https://kojipkgs.fedoraproject.org/compose/Fedora-Modular-27-20171115.n.1/'
        mapping = {
            # we just pretend 1 is 6, it's easier
            arcbase + 'core/1/SRPMS/': FC6PACKAGEHTML,
            arcbase + 'core/6/source/SRPMS/': FC6PACKAGEHTML,
            arcbase + 'releases/16/Everything/source/SRPMS/': F16PACKAGEHTML,
            arcbase + 'releases/18/Everything/source/SRPMS/a/': F18PACKAGEHTMLA,
            arcbase + 'releases/18/Everything/source/SRPMS/b/': F18PACKAGEHTMLB,
            arcbase + 'releases/18/Everything/source/SRPMS/f/': '',
            # similarly we just use the 18 content for 24 and mod27
            branchbase + 'compose/Everything/source/tree/Packages/a/': F18PACKAGEHTMLA,
            branchbase + 'compose/Everything/source/tree/Packages/b/': F18PACKAGEHTMLB,
            branchbase + 'compose/Everything/source/tree/Packages/f/': '',
            modbase + 'compose/Server/source/tree/Packages/a/': F18PACKAGEHTMLA,
            modbase + 'compose/Server/source/tree/Packages/b/': F18PACKAGEHTMLB,
            modbase + 'compose/Server/source/tree/Packages/f/': '',
        }
        if self.url in mapping:
            return six.b(mapping[self.url])
        else:
            raise ValueError("URL {0} not expected!".format(self.url))

def urlopen_fake_package(url):
    """Stub for urlopen_retries which returns FakeResponsePackages."""
    return FakeResponsePackages(url)

@pytest.mark.usefixtures("clean_home")
class TestRelease:
    """Tests for release.py."""

    @pytest.mark.parametrize(
        ("release", "milestone", "compose", "respin", "dist", "expclass",
         "exprel", "expmile", "expcomp", "expspin"),
        # rel/mile/comp/respin to pass, expected class, expected rel/mile/comp/respin
        # None indicates check should not be done
        [
            # Archive stable.
            (15, '', '', None, "Fedora", fedfind.release.ArchiveRelease, '15', None, None, None),
            # Fedora Core stable (release string).
            ('6', '', '', None, "Fedora", fedfind.release.CoreRelease, '6', None, None, None),
            # 'Final' treated as stable.
            ('6', 'Final', '', None, "Fedora", fedfind.release.CoreRelease, '6', None, None, None),
            # Milestone.
            (23, 'Beta', '', None, "Fedora", fedfind.release.Milestone, '23', 'Beta', None, None),
            # Branched nightly.
            (23, 'Branched', 'datestr', 1, "Fedora",
             fedfind.release.BranchedNightly, '23', 'Branched', 'datestr', '1'),
            # Rawhide nightly.
            ('Rawhide', '', 'datestr', 2, "Fedora",
             fedfind.release.RawhideNightly, 'Rawhide', '', 'datestr', '2'),
            # Atomic nightly (old, milestone way).
            (24, 'Atomic', 'datestr', 0, "Fedora",
             fedfind.release.AtomicNightly, '24', 'Nightly', 'datestr', '0'),
            # Docker nightly (old, milestone way).
            (25, 'Docker', 'datestr', 0, "Fedora",
             fedfind.release.DockerNightly, '25', 'Nightly', 'datestr', '0'),
            # Cloud nightly (old, milestone way).
            (25, 'Cloud', 'datestr', 0, "Fedora",
             fedfind.release.CloudNightly, '25', 'Nightly', 'datestr', '0'),
            # Atomic nightly (new, dist way).
            (24, '', 'datestr', 0, "Fedora-Atomic",
             fedfind.release.AtomicNightly, '24', 'Nightly', 'datestr', '0'),
            # Docker nightly (new, dist way).
            (25, '', 'datestr', 0, "Fedora-Docker",
             fedfind.release.DockerNightly, '25', 'Nightly', 'datestr', '0'),
            # Cloud nightly (new, dist way).
            (25, '', 'datestr', 0, "Fedora-Cloud",
             fedfind.release.CloudNightly, '25', 'Nightly', 'datestr', '0'),
            # Wikitcms format Rawhide.
            ('24', 'Rawhide', 'datestr.n.0', None, "Fedora",
             fedfind.release.RawhideNightly, 'Rawhide', '', 'datestr', '0'),
            # OFFLINE GUESSES
            # Date guesses.
            (23, 'Branched', '', 0, "Fedora",
             fedfind.release.BranchedNightly, '23', 'Branched', 'datestr', '0'),
            (27, 'Branched', '', 0, "Fedora-Modular",
             fedfind.release.BranchedModularNightly, '27', 'Branched', 'datestr', '0'),
            ('Rawhide', '', '', 0, "Fedora",
             fedfind.release.RawhideNightly, 'Rawhide', '', 'datestr', '0'),
            ('Rawhide', '', '', 0, "Fedora-Modular",
             fedfind.release.RawhideModularNightly, 'Rawhide', '', 'datestr', '0'),
            ('Bikeshed', '', '', 0, "Fedora-Modular",
             fedfind.release.BikeshedModularNightly, 'Bikeshed', '', 'datestr', '0'),
            (24, 'Atomic', '', 0, "Fedora",
             fedfind.release.AtomicNightly, '24', 'Nightly', 'datestr', '0'),
            (25, 'Docker', '', 0, "Fedora",
             fedfind.release.DockerNightly, '25', 'Nightly', 'datestr', '0'),
            (25, 'Cloud', '', 0, "Fedora",
             fedfind.release.CloudNightly, '25', 'Nightly', 'datestr', '0'),
            # Default, check None is handled.
            (None, None, None, 0, "Fedora",
             fedfind.release.RawhideNightly, 'Rawhide', '', 'datestr', '0'),
        ]
    )
    # pylint:disable=too-many-arguments
    def test_get_release_simple(self, release, milestone, compose, respin, dist, expclass, exprel,
                                expmile, expcomp, expspin):
        """Tests for get_release that require no online guessing or
        checking, hence no mocking.
        """
        datestr = datetime.date.today().strftime('%Y%m%d')
        if compose:
            compose = compose.replace('datestr', datestr)
        if expcomp:
            expcomp = expcomp.replace('datestr', datestr)
        got = fedfind.release.get_release(release, milestone, compose, respin, dist=dist)
        assert isinstance(got, expclass)
        if exprel is not None:
            assert got.release == exprel
        if expmile is not None:
            assert got.milestone == expmile
        if expcomp is not None:
            assert got.compose == expcomp
        if expspin is not None:
            assert got.respin == expspin

        # Sanity check tests.
        with pytest.raises(ValueError):
            fedfind.release.get_release('foobar')

        with pytest.raises(ValueError):
            fedfind.release.get_release('23', '', 'RC2')

    @mock.patch('fedfind.release.Production.label', 'Alpha-1.1')
    def test_get_release_production(self):
        """Explicitly getting a Production (i.e. a production compose
        on kojipkgs, not mirrored to alt, which is a Compose). This
        is the same path hit if get_release is called with a compose
        ID for a production compose, and promote is not set True.
        """
        ret = fedfind.release.get_release(24, 'Production', '20160314', 0)
        assert isinstance(ret, fedfind.release.Production)
        assert ret.release == '24'
        assert ret.milestone == 'Alpha'
        assert ret.compose == '1'
        assert ret.respin == '1'

    @mock.patch('fedfind.helpers.label_from_cid', return_value="Alpha-1.1", autospec=True)
    @mock.patch('fedfind.release.Compose.exists', True)
    def test_get_release_cid_promote(self, fakelab):
        """Requesting release from a production compose ID with
        promote set to True should return a Compose if possible.
        """
        ret = fedfind.release.get_release(cid='Fedora-24-20160314.0', promote=True)
        assert isinstance(ret, fedfind.release.Compose)
        assert ret.release == '24'
        assert ret.milestone == 'Alpha'
        assert ret.compose == '1'
        assert ret.respin == '1'
        fakelab.assert_called_with('Fedora-24-20160314.0')

    @mock.patch('fedfind.helpers.label_from_cid', return_value="Alpha-1.1", autospec=True)
    @mock.patch('fedfind.helpers.cid_from_label', autospec=True)
    @mock.patch('fedfind.release.Compose.exists', False)
    # this is to avoid a network trip for the get_release CID check
    @mock.patch('fedfind.release.Production.cid', 'Fedora-24-20160314.0')
    def test_get_release_cid_promote_fail(self, fakecid, fakelab):
        """This tests a rather snaky path: we call get_release with
        a CID and promote=True; the specified compose has a label, so
        we try and get a Compose, but the compose is not mirrored, so
        we fail. In this case we should fall back on a Production *and
        not waste time on a remote trip to recreate the CID*, instead
        it should be reused.
        """
        ret = fedfind.release.get_release(cid='Fedora-24-20160314.0', promote=True)
        assert isinstance(ret, fedfind.release.Production)
        fakelab.assert_called_with('Fedora-24-20160314.0')
        assert fakecid.call_count == 0

    @mock.patch('fedfind.release.ModularProduction.exists', True)
    @mock.patch('fedfind.release.ModularProduction.cid', 'Fedora-Modular-27-20171107.1')
    def test_get_release_modular_production_cid(self):
        """Test getting a ModularProduction instance via compose ID."""
        ret = fedfind.release.get_release(cid='Fedora-Modular-27-20171107.1')
        assert isinstance(ret, fedfind.release.ModularProduction)

    @mock.patch('fedfind.release.ModularCompose.exists', True)
    def test_get_release_modular_compose(self):
        """Test getting a ModularCompose instance."""
        ret = fedfind.release.get_release(27, 'Beta', 1, 1, dist="Fedora-Modular")
        assert isinstance(ret, fedfind.release.ModularCompose)

    @mock.patch('fedfind.release.AtomicNightly.cid', 'Fedora-Atomic-24-20160628.1')
    def test_get_release_cid_atomic_new(self):
        """Test that nightly stable Atomic compose IDs are handled.
        They have 'Fedora-Atomic' as the product name instead of
        'Fedora'.
        """
        ret = fedfind.release.get_release(cid='Fedora-Atomic-24-20160628.1')
        assert isinstance(ret, fedfind.release.AtomicNightly)
        assert ret.release == '24'
        assert ret.compose == '20160628'
        assert ret.respin == '1'

    @mock.patch('fedfind.release.DockerNightly.cid', 'Fedora-Docker-25-20170113.2')
    def test_get_release_cid_docker(self):
        """Test that nightly stable Docker compose IDs are handled.
        They have 'Fedora-Docker' as the product name instead of
        'Fedora'.
        """
        ret = fedfind.release.get_release(cid='Fedora-Docker-25-20170113.2')
        assert isinstance(ret, fedfind.release.DockerNightly)
        assert ret.release == '25'
        assert ret.compose == '20170113'
        assert ret.respin == '2'

    @mock.patch('fedfind.release.CloudNightly.cid', 'Fedora-Cloud-26-20170217.3')
    def test_get_release_cid_cloud(self):
        """Test that nightly stable Cloud compose IDs are handled.
        They have 'Fedora-Cloud' as the product name instead of
        'Fedora'.
        """
        ret = fedfind.release.get_release(cid='Fedora-Cloud-26-20170217.3')
        assert isinstance(ret, fedfind.release.CloudNightly)
        assert ret.release == '26'
        assert ret.compose == '20170217'
        assert ret.respin == '3'

    def test_get_release_cid_fake_stable(self):
        """Getting a release based on the fake CID produced for stable
        releases should return the same release.
        """
        rel = fedfind.release.get_release(15)
        rel2 = fedfind.release.get_release(cid=rel.cid)
        assert rel2.version == rel.version

    def test_get_release_url_cid_rawhide(self):
        """Getting a release based on a correct Rawhide nightly URL
        should work.
        """
        url = ('https://kojipkgs.fedoraproject.org/'
               'compose/rawhide/Fedora-Rawhide-20170126.n.0/compose/')
        ret = fedfind.release.get_release(url=url)
        assert isinstance(ret, fedfind.release.RawhideNightly)
        assert ret.release == 'Rawhide'
        assert ret.compose == '20170126'
        assert ret.respin == '0'

    def test_get_release_url_cid_branched(self):
        """Getting a release based on a correct Branched nightly URL
        should work.
        """
        url = ('https://kojipkgs.fedoraproject.org/'
               'compose/branched/Fedora-27-20170816.n.0/compose/')
        ret = fedfind.release.get_release(url=url)
        assert isinstance(ret, fedfind.release.BranchedNightly)
        assert ret.release == '27'
        assert ret.compose == '20170816'
        assert ret.respin == '0'

    def test_get_release_url_cid_rawhide_modular(self):
        """Getting a release based on a correct Rawhide modular
        nightly URL should work.
        """
        url = ('https://kojipkgs.fedoraproject.org/'
               'compose/Fedora-Modular-Rawhide-20170816.n.0/compose/')
        ret = fedfind.release.get_release(url=url)
        assert isinstance(ret, fedfind.release.RawhideModularNightly)
        assert ret.release == 'Rawhide'
        assert ret.compose == '20170816'
        assert ret.respin == '0'

    def test_get_release_url_cid_bikeshed_modular(self):
        """Getting a release based on a correct Bikeshed modular
        nightly URL should work.
        """
        url = ('https://kojipkgs.fedoraproject.org/'
               'compose/Fedora-Modular-Bikeshed-20171013.n.0/compose/')
        ret = fedfind.release.get_release(url=url)
        assert isinstance(ret, fedfind.release.BikeshedModularNightly)
        assert ret.release == 'Bikeshed'
        assert ret.compose == '20171013'
        assert ret.respin == '0'

    def test_get_release_url_cid_branched_modular(self):
        """Getting a release based on a correct Branched modular
        nightly URL should work.
        """
        url = ('https://kojipkgs.fedoraproject.org/'
               'compose/Fedora-Modular-27-20170816.n.0/compose/')
        ret = fedfind.release.get_release(url=url)
        assert isinstance(ret, fedfind.release.BranchedModularNightly)
        assert ret.release == '27'
        assert ret.compose == '20170816'
        assert ret.respin == '0'

    def test_get_release_url_cid_unknown(self):
        """Getting a release based on an unknown URL should return
        a generic Pungi4Release instance, so long as the URL starts
        with an acceptable string (to avoid puiterwijk's 'evil URL
        input' scenario).
        """
        ret = fedfind.release.get_release(
            url='https://kojipkgs.fedoraproject.org/compose/whatthehellisthis/ihavenoidea/compose/')
        assert isinstance(ret, fedfind.release.Pungi4Release)

    def test_get_release_url_cid_unknown_evil(self):
        """Getting a release based on an unknown URL that doesn't
        start with one of our known prefixes should raise a value
        error.
        """
        url = 'https://foobar.fedoraproject.org/compose/whatthehellisthis/ihavenoidea/compose/'
        with pytest.raises(ValueError):
            fedfind.release.get_release(url=url)

    def test_get_release_url_check(self):
        """Test the URL match check for get_release: if we get a
        release by URL, but the URL of the discovered Release instance
        doesn't match the requested URL, an exception should occur.
        """
        with pytest.raises(fedfind.exceptions.UrlMatchError):
            url = 'https://kojipkgs.fedoraproject.org/wrong/Fedora-Rawhide-20180117.n.0/compose/'
            fedfind.release.get_release(url=url)

    # This would be the correct compose ID value for a Rawhide nightly
    @mock.patch('fedfind.release.RawhideNightly.cid', 'Fedora-Rawhide-20180117.n.0')
    def test_get_release_cid_check(self):
        """Test the CID match check for get_release: if we get a
        release by CID, but the CID of the discovered Release instance
        doesn't match the requested CID, an exception should occur.
        """
        with pytest.raises(fedfind.exceptions.CidMatchError):
            # real compose IDs for Rawhide are always '.n.', as we
            # patch above; get_release will successfully handle this
            # CID, though, and give us a RawhideNightly instance
            cid = 'Fedora-Rawhide-20180117.0'
            fedfind.release.get_release(cid=cid)

    @mock.patch('fedfind.release.Compose.exists', True)
    def test_get_release_compose(self):
        """A production/candidate compose. This tests the case where
        the compose is found on the mirror system. Also tests that
        'Final' is converted to 'RC'.
        """
        ret = fedfind.release.get_release(24, 'Final', 1, 1)
        assert isinstance(ret, fedfind.release.Compose)
        assert ret.release == '24'
        assert ret.milestone == 'RC'
        assert ret.compose == '1'
        assert ret.respin == '1'
        assert ret.label == 'RC-1.1'

    @mock.patch('fedfind.helpers.cid_from_label', return_value='Fedora-24-20160316.3',
                autospec=True)
    @mock.patch('fedfind.release.Compose.exists', False)
    def test_get_release_compose_fallback(self, fakecid):
        """A production/candidate compose. This tests the case where
        the compose is not found on the mirror system so we fall back
        on finding the compose ID and returning a Production.
        """
        ret = fedfind.release.get_release(24, 'Alpha', 1, 5)
        assert isinstance(ret, fedfind.release.Production)
        fakecid.assert_called_with(24, 'Alpha-1.5')

    @mock.patch('fedfind.release.CurrentRelease.exists', True)
    def test_get_stable_current(self):
        """Ensure that if it exists, get_release for a recent stable
        returns CurrentRelease. Tested number must be higher than the
        'just return ArchiveRelease' cutoff, so bump this test when
        bumping that cutoff.
        """
        ret = fedfind.release.get_release(26)
        assert isinstance(ret, fedfind.release.CurrentRelease)
        assert ret.release == '26'

    @mock.patch('fedfind.release.CurrentRelease.exists', False)
    @mock.patch('fedfind.release.ArchiveRelease.exists', True)
    def test_get_stable_archive(self):
        """Ensure that if CurrentRelease does not exist but
        ArchiveRelease does exist, get_release for a recent stable
        returns it. Tested number must be higher than the 'just return
        ArchiveRelease' cutoff (or else this test will pass even if
        the feature is broken), so bump this test when bumping that
        cutoff.
        """
        ret = fedfind.release.get_release(26)
        assert isinstance(ret, fedfind.release.ArchiveRelease)
        assert ret.release == '26'

    @mock.patch('fedfind.helpers.get_current_release', return_value=23, autospec=True)
    def test_get_release_guess_release(self, fakecurrent):
        """Release guessing. We can't test this very hard or we'd have
        to keep updating the tests all the goddamned time, or the test
        would duplicate the logic of get_current_release and that
        seems pointless. But we can at least check it's not crashing
        and returns the right class.
        """
        datestr = datetime.date.today().strftime('%Y%m%d')

        got = fedfind.release.get_release('', 'Branched', '', 0)
        assert isinstance(got, fedfind.release.BranchedNightly)
        assert got.compose == datestr
        assert got.release == '24'

        # old 'milestone' style
        got = fedfind.release.get_release('', 'Atomic', '', 0)
        assert isinstance(got, fedfind.release.AtomicNightly)
        assert got.compose == datestr
        assert got.release == '23'

        # old 'milestone' style
        got = fedfind.release.get_release('', 'Docker', '', 0)
        assert isinstance(got, fedfind.release.DockerNightly)
        assert got.compose == datestr
        assert got.release == '23'

        # old 'milestone' style
        got = fedfind.release.get_release('', 'Cloud', '', 0)
        assert isinstance(got, fedfind.release.CloudNightly)
        assert got.compose == datestr
        assert got.release == '23'

        # new 'dist' style
        got = fedfind.release.get_release('', '', '', 0, dist='Fedora-Atomic')
        assert isinstance(got, fedfind.release.AtomicNightly)
        assert got.compose == datestr
        assert got.release == '23'

        # new 'dist' style
        got = fedfind.release.get_release('', '', '', 0, dist='Fedora-Docker')
        assert isinstance(got, fedfind.release.DockerNightly)
        assert got.compose == datestr
        assert got.release == '23'

        # new 'dist' style
        got = fedfind.release.get_release('', '', '', 0, dist='Fedora-Cloud')
        assert isinstance(got, fedfind.release.CloudNightly)
        assert got.compose == datestr
        assert got.release == '23'

    @mock.patch('fedfind.release.RawhideNightly', FakeRelease)
    @mock.patch('fedfind.release.RawhideModularNightly', FakeRelease)
    @mock.patch('fedfind.release.BikeshedModularNightly', FakeRelease)
    @mock.patch('fedfind.release.BranchedNightly', FakeRelease)
    @mock.patch('fedfind.release.BranchedModularNightly', FakeRelease)
    @mock.patch('fedfind.release.AtomicNightly', FakeRelease)
    @mock.patch('fedfind.release.DockerNightly', FakeRelease)
    @mock.patch('fedfind.release.CloudNightly', FakeRelease)
    def test_get_release_guess_respin(self):
        """Test the 'respin guessing' code works (when you request a
        nightly but don't provide a respin number). Uses the fake
        release class's feature of 'existing' when the respin is 3.
        """
        got = fedfind.release.get_release(24, 'Branched', '20160314')
        assert isinstance(got, fedfind.release.BranchedNightly)
        assert got.respin == 3
        got = fedfind.release.get_release(27, 'Branched', '20170816', dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.BranchedModularNightly)
        assert got.respin == 3
        got = fedfind.release.get_release('Rawhide', '', '20160314')
        assert isinstance(got, fedfind.release.RawhideNightly)
        assert got.respin == 3
        got = fedfind.release.get_release('Rawhide', '', '20170816', dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.RawhideModularNightly)
        assert got.respin == 3
        got = fedfind.release.get_release('Bikeshed', '', '20171013', dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.BikeshedModularNightly)
        assert got.respin == 3
        # old 'milestone' style
        got = fedfind.release.get_release(24, 'Atomic', '20160628')
        assert isinstance(got, fedfind.release.AtomicNightly)
        assert got.respin == 3
        got = fedfind.release.get_release(24, 'Docker', '20160628')
        assert isinstance(got, fedfind.release.DockerNightly)
        assert got.respin == 3
        got = fedfind.release.get_release(24, 'Cloud', '20160628')
        assert isinstance(got, fedfind.release.CloudNightly)
        assert got.respin == 3
        # new 'dist' style
        got = fedfind.release.get_release(24, '', '20160628', dist='Fedora-Atomic')
        assert isinstance(got, fedfind.release.AtomicNightly)
        assert got.respin == 3
        got = fedfind.release.get_release(24, '', '20160628', dist='Fedora-Docker')
        assert isinstance(got, fedfind.release.DockerNightly)
        assert got.respin == 3
        got = fedfind.release.get_release(24, '', '20160628', dist='Fedora-Cloud')
        assert isinstance(got, fedfind.release.CloudNightly)
        assert got.respin == 3

    @mock.patch.object(fedfind.release.BranchedNightly, 'exists', True)
    @mock.patch.object(fedfind.release.BranchedNightly, 'status', 'FINISHED')
    @mock.patch.object(fedfind.release.BranchedModularNightly, 'exists', True)
    @mock.patch.object(fedfind.release.BranchedModularNightly, 'status', 'FINISHED')
    @mock.patch.object(fedfind.release.AtomicNightly, 'exists', True)
    @mock.patch.object(fedfind.release.AtomicNightly, 'status', 'FINISHED')
    def test_get_release_guess_nightly_branched_exists(self):
        """Test nightly type guessing: if we specify a date and a
        release number, but no milestone, fedfind should guess between
        BranchedNightly and AtomicNightly. If Branched exists, we
        should return it. Also check this path works OK for modular
        Branched.
        """
        got = fedfind.release.get_release(25, '', '20161006')
        assert isinstance(got, fedfind.release.BranchedNightly)
        got = fedfind.release.get_release(25, '', '20161006', 3)
        assert isinstance(got, fedfind.release.BranchedNightly)
        got = fedfind.release.get_release(27, '', '20170816', dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.BranchedModularNightly)
        got = fedfind.release.get_release(27, '', '20170816', 3, dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.BranchedModularNightly)

    @mock.patch.object(fedfind.release.BranchedNightly, 'exists', False)
    @mock.patch.object(fedfind.release.BranchedNightly, 'status', 'DOOMED')
    @mock.patch.object(fedfind.release.BranchedModularNightly, 'exists', False)
    @mock.patch.object(fedfind.release.BranchedModularNightly, 'status', 'DOOMED')
    @mock.patch.object(fedfind.release.AtomicNightly, 'exists', True)
    @mock.patch.object(fedfind.release.AtomicNightly, 'status', 'FINISHED')
    def test_get_release_guess_nightly_branched_notexists(self):
        """Test nightly type guessing: if we specify a date and a
        release number, but no milestone, fedfind should guess between
        BranchedNightly and AtomicNightly. If Branched doesn't exist
        but Atomic does, we should return Atomic. Also check this path
        works OK for modular Branched.
        """
        got = fedfind.release.get_release(25, '', '20161006')
        assert isinstance(got, fedfind.release.AtomicNightly)
        got = fedfind.release.get_release(25, '', '20161006', 3)
        assert isinstance(got, fedfind.release.AtomicNightly)
        got = fedfind.release.get_release(27, '', '20170816', dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.AtomicNightly)
        got = fedfind.release.get_release(27, '', '20170816', 3, dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.AtomicNightly)

    @mock.patch.object(fedfind.release.BranchedNightly, 'exists', False)
    @mock.patch.object(fedfind.release.BranchedNightly, 'status', 'DOOMED')
    @mock.patch.object(fedfind.release.BranchedModularNightly, 'exists', False)
    @mock.patch.object(fedfind.release.BranchedModularNightly, 'status', 'DOOMED')
    @mock.patch.object(fedfind.release.AtomicNightly, 'exists', False)
    @mock.patch.object(fedfind.release.AtomicNightly, 'status', 'DOOMED')
    def test_get_release_guess_nightly_neitherexists(self):
        """Test nightly type guessing: if we specify a date and a
        release number, but no milestone, fedfind should guess between
        BranchedNightly and AtomicNightly. If neither exists, we should
        return Branched.
        """
        got = fedfind.release.get_release(25, '', '20161006')
        assert isinstance(got, fedfind.release.BranchedNightly)
        got = fedfind.release.get_release(25, '', '20161006', 3)
        assert isinstance(got, fedfind.release.BranchedNightly)
        got = fedfind.release.get_release(27, '', '20170816', dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.BranchedModularNightly)
        got = fedfind.release.get_release(27, '', '20170816', 3, dist='Fedora-Modular')
        assert isinstance(got, fedfind.release.BranchedModularNightly)

    def test_get_release_respinrelease(self, http):
        """Test get_release for RespinRelease class. We use the fake
        web server here as RespinRelease has to discover its own
        release and compose from the image list.
        """
        # old milestone style
        got = fedfind.release.get_release('', 'Respin', '')
        assert isinstance(got, fedfind.release.RespinRelease)
        assert got.release == '27'
        assert got.milestone == 'Respin'
        assert got.compose == '20180112'
        # getting a release from the fake compose ID should get the
        # same release
        got2 = fedfind.release.get_release(cid=got.cid)
        assert got.version == got2.version
        # getting a release with the old form of fake compose ID also
        got3 = fedfind.release.get_release(cid='FedoraRespin-27-20180112.0')
        assert got.version == got3.version
        # getting a release from the URL should get the same release
        got4 = fedfind.release.get_release(
            url='https://dl.fedoraproject.org/pub/alt/live-respins/')
        assert got4.version == got3.version
        # new dist style (should match others)
        got5 = fedfind.release.get_release('', '', '', dist='FedoraRespin')
        assert got5.version == got4.version
        # trying to get RespinRelease with a non-matching release
        # or compose should raise an error
        with pytest.raises(ValueError):
            fedfind.release.get_release('25', 'Respin', '')
        with pytest.raises(ValueError):
            fedfind.release.get_release('', 'Respin', '20170305')
        with pytest.raises(ValueError):
            fedfind.release.get_release('25', 'Respin', '20170305')
        with pytest.raises(ValueError):
            fedfind.release.get_release(cid='FedoraRespin-25-updates-20170305.0')

    @mock.patch('fedfind.helpers.urlopen_retries', urlopen_fake_package)
    def test_get_package_nvras_fc1(self):
        """FC1: no /Everything, no /source, no split-by-initials."""
        rel = fedfind.release.get_release(1)
        pkgs = rel.get_package_nvras(['amanda', 'anaconda', 'bash', 'fakepackage'])
        assert pkgs == {
            'amanda': 'amanda-2.5.0p2-4.src',
            'anaconda': 'anaconda-11.1.1.3-1.src',
            'bash': 'bash-3.1-16.1.src',
            'fakepackage': ''
        }

    @mock.patch('fedfind.helpers.urlopen_retries', urlopen_fake_package)
    def test_get_package_nvras_fc6(self):
        """FC6: Still no /Everything, /source appeared, no
        split-by-initials.
        """
        rel = fedfind.release.get_release(6)
        pkgs = rel.get_package_nvras(['amanda', 'anaconda', 'bash', 'fakepackage'])
        assert pkgs == {
            'amanda': 'amanda-2.5.0p2-4.src',
            'anaconda': 'anaconda-11.1.1.3-1.src',
            'bash': 'bash-3.1-16.1.src',
            'fakepackage': ''
        }

    @mock.patch('fedfind.helpers.urlopen_retries', urlopen_fake_package)
    def test_get_package_nvras_f16(self):
        """F16: /Everything showed up, still no split-by-initials."""
        rel = fedfind.release.get_release(16)
        pkgs = rel.get_package_nvras(['amanda', 'anaconda', 'bash', 'fakepackage'])
        assert pkgs == {
            'amanda': 'amanda-3.3.0-2.fc16.src',
            'anaconda': 'anaconda-16.25-1.fc16.src',
            'bash': 'bash-4.2.10-4.fc16.src',
            'fakepackage': ''
        }

    @mock.patch('fedfind.helpers.urlopen_retries', urlopen_fake_package)
    def test_get_package_nvras_f18(self):
        """F18: split-by-initials now in effect."""
        rel = fedfind.release.get_release(18)
        pkgs = rel.get_package_nvras(['amanda', 'anaconda', 'bash', 'fakepackage'])
        assert pkgs == {
            'amanda': 'amanda-3.3.2-2.fc18.src',
            'anaconda': 'anaconda-18.37.11-1.fc18.src',
            'bash': 'bash-4.2.39-3.fc18.src',
            'fakepackage': ''
        }

#    @mock.patch('fedfind.helpers.pdc_query', return_value=PDCPACKAGES)
    @mock.patch('fedfind.helpers.urlopen_retries', urlopen_fake_package)
    @mock.patch('fedfind.release.BranchedNightly.cid', "Fedora-24-20160321.n.0")
    def test_get_package_nvras_f24(self):
#    def test_get_package_nvras_f24(self, fakepdc):
        """F24: Different layout again under /Everything. mock values
        are same as for F18. Commented-out lines are for testing the
        PDC version of the method, which is currently disabled: see
        comment in Pungi4Release class.
        """
        rel = fedfind.release.get_release(24, 'Branched', '20160321', 0)
        pkgs = rel.get_package_nvras(['amanda', 'anaconda', 'bash', 'fakepackage'])
        assert pkgs == {
            'amanda': 'amanda-3.3.2-2.fc18.src',
            'anaconda': 'anaconda-18.37.11-1.fc18.src',
            'bash': 'bash-4.2.39-3.fc18.src',
            'fakepackage': ''
        }
#        assert pkgs == {
#            'amanda': 'amanda-0:3.3.8-1.fc24.src',
#            'anaconda': 'anaconda-0:24.13-1.fc24.src',
#            'bash': 'bash-0:4.3.42-4.fc24.src',
#        }
#        fakepdc.assert_called_with(
#            'rpms',
#            [
#                ('compose', 'Fedora-24-20160321.n.0'),
#                ('arch', 'src'),
#                ('name', 'amanda'),
#                ('name', 'anaconda'),
#                ('name', 'bash'),
#                ('name', 'fakepackage')
#            ]
#        )

#    @mock.patch('fedfind.helpers.pdc_query', return_value=PDCPACKAGES)
    @mock.patch('fedfind.helpers.urlopen_retries', urlopen_fake_package)
    @mock.patch('fedfind.release.BranchedModularNightly.cid', 'Fedora-Modular-27-20171115.n.1')
    def test_get_package_nvras_mod_f27(self):
#    def test_get_package_nvras_mod_f27(self, fakepdc):
        """Modular F27: No /Everything tree, only /Server. mock values
        are same as for F18. Commented-out lines are for testing the
        PDC version of the method, which is currently disabled: see
        comment in Pungi4Release class.
        """
        rel = fedfind.release.get_release(cid="Fedora-Modular-27-20171115.n.1")
        pkgs = rel.get_package_nvras(['amanda', 'anaconda', 'bash', 'fakepackage'])
        assert pkgs == {
            'amanda': 'amanda-3.3.2-2.fc18.src',
            'anaconda': 'anaconda-18.37.11-1.fc18.src',
            'bash': 'bash-4.2.39-3.fc18.src',
            'fakepackage': ''
        }
#        assert pkgs == {
#            'amanda': 'amanda-0:3.3.8-1.fc24.src',
#            'anaconda': 'anaconda-0:24.13-1.fc24.src',
#            'bash': 'bash-0:4.3.42-4.fc24.src',
#        }
#        fakepdc.assert_called_with(
#            'rpms',
#            [
#                ('compose', 'Fedora-24-20160321.n.0'),
#                ('arch', 'src'),
#                ('name', 'amanda'),
#                ('name', 'anaconda'),
#                ('name', 'bash'),
#                ('name', 'fakepackage')
#            ]
#        )

    def test_all_stable(self, http, pdc):
        """Test that we get the expected image dicts for all stable
        releases, using the test HTTP server and home directory and
        the PDC query mock. Also check there are no missing expected
        images (as a test for check_expected). Note this test is
        # sensitive to the stupid 'current/archive' detection in
        # get_release: it will fail if the hardcoded cutoff there
        # doesn't match the data in allstable.json.
        """
        ref = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'allstable.json')
        with codecs.open(ref, encoding='utf-8') as reffh:
            expected = json.loads(reffh.read())
        # one thing we can handily check here, each time we bump the
        # backing data for this test to include a new release, is if
        # any sneaky new subvariants showed up...
        subvs = [subv.name for subv in fedfind.const.SUBVARIANTS]
        # for some bizarre reason, if we do this as a decorator like
        # normal, the test blows up on Python 2.6
        patcher = mock.patch('fedfind.release.CurrentRelease.exists', True)
        patcher.start()
        for relnum in expected:
            rel = fedfind.release.get_release(relnum)
            imgs = sorted(rel.all_images, key=lambda x: x['path'])
            assert expected[relnum] == imgs
            for img in imgs:
                subv = img.get('subvariant')
                if subv:
                    assert subv in subvs
            # we have to force the release to 'exist' or else
            # check_expected bails
            rel._exists = True
            assert not rel.check_expected()
        patcher.stop()

    def test_images_respinrelease(self, http):
        """Test that we get approximately the expected image dicts for
        RespinRelease. This depends on the contents of imagelist-alt
        in the fake http server; if that's updated, this test may have
        to change. Especially note that the 20161120 respin didn't
        have an SOAS image.
        """
        rel = fedfind.release.RespinRelease()
        imgs = rel.all_images
        assert len(imgs) == 8
        # now ditch source, as we don't care much about it
        imgs = [img for img in imgs if img['subvariant'] != 'Source']
        assert len(imgs) == 7
        assert all(img['type'] == 'live' for img in imgs)
        # this is a lie, but it's the lie we decided to tell
        assert all(img['variant'] == 'Spins' for img in imgs)
        assert all(img['arch'] == 'x86_64' for img in imgs)
        exp = ['Cinnamon', 'KDE', 'LXDE', 'LXQt', 'Mate', 'Workstation', 'Xfce']
        got = sorted([img['subvariant'] for img in imgs])
        assert got == exp

    def test_compose_urls(self, http):
        """Check that get_release(url=) works sanely for a list of 180
        days worth of compose URLs. List is generated by fedfindloc.py
        script.
        """
        urlfn = 'compose-urls-20180112.json'
        urlfp = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', urlfn)
        with codecs.open(urlfp, encoding='utf-8') as urlfh:
            urls = json.loads(urlfh.read())
        for url in urls:
            try:
                assert fedfind.release.get_release(url=url)
            except fedfind.exceptions.UnsupportedComposeError:
                pass
            except fedfind.exceptions.UrlMatchError as err:
                # A couple of really pathological compose ID types we
                # managed to come up with which I'm not interested in
                # teaching the compose ID parser about:
                if "Fedora-Modular-f27-updates-Fedora-Modular" in err.url:
                    pass
                elif "Fedora-Modular-f27-updates-updates-Fedora-Modular" in err.url:
                    pass
                # these exist in both compose/ and compose/twoweek/ for some
                # reason, doesn't seem worth handling
                elif "compose/Fedora-Atomic-26-201707" in err.url:
                    pass
                else:
                    raise

# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) Red Hat Inc.
#
# autocloudreporter is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:   Adam Williamson <awilliam@redhat.com>

"""Setuptools script."""

import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """Pytest integration."""
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''
        self.test_suite = 'tests'

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args.split())
        sys.exit(errno)


# From: https://github.com/pypa/pypi-legacy/issues/148
# Produce rst-formatted long_description if pypandoc is available (to
# look nice on pypi), otherwise just use the Markdown-formatted one
try:
    import pypandoc
    longdesc = pypandoc.convert('README.md', 'rst')
except ImportError:
    longdesc = open('README.md').read()

setup(
    name="fedfind",
    version="4.0.0",
    entry_points={'console_scripts': ['fedfind = fedfind.cli:main'],},
    author="Adam Williamson",
    author_email="awilliam@redhat.com",
    description="Fedora Finder finds Fedora - images, more to come?",
    license="GPLv3+",
    keywords="fedora release image media iso",
    url="https://pagure.io/fedora-qa/fedfind",
    packages=["fedfind"],
    install_requires=open('install.requires').read().splitlines(),
    tests_require=open('tests.requires').read().splitlines(),
    cmdclass={'test': PyTest},
    long_description=longdesc,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License v3 or later "
        "(GPLv3+)",
    ],
)

# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2015 Red Hat
#
# This file is part of fedfind.
#
# fedfind is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""CLI module for fedfind."""

from __future__ import unicode_literals
from __future__ import print_function

import argparse
import logging
import sys

import fedfind.const
import fedfind.exceptions
import fedfind.helpers
import fedfind.release

# pylint:disable=invalid-name
logger = logging.getLogger(__name__)

def _release(rel):
    """Validity check for release parameter - must be an integer or
    rawhide.
    """
    if rel.isdigit() or rel.lower() == 'rawhide':
        return rel
    else:
        raise ValueError("Release must be integer or Rawhide.")

def images(args):
    """images sub-command function. Gets an appropriate Release
    instance and runs an appropriate query, based on the provided
    args.
    """
    # pylint:disable=too-many-branches
    try:
        rel = fedfind.release.get_release(
            dist=args.dist, release=args.release, milestone=args.milestone, compose=args.compose,
            respin=args.respin, cid=args.composeid, label=args.label)
        logger.debug("Release instance: %s", rel)
        if isinstance(rel, fedfind.release.Pungi4Release):
            if not rel.status in fedfind.const.PUNGI_SUCCESS:
                sys.exit("Compose {0} failed, or not yet complete!".format(rel))
        logger.info("Finding images for: %s %s", rel.dist, rel.version)
    except fedfind.exceptions.CidMatchError as err:
        sys.exit("Compose ID {0} of discovered compose at {1} does not match specified compose ID "
                 "{2}! Please report a bug.".format(err.release.cid, err.release.location,
                                                    args.composeid))
    except fedfind.exceptions.UnsupportedComposeError as err:
        sys.exit("Compose is unsupported. {0}".format(str(err)))
    except ValueError as err:
        sys.exit("Invalid arguments. {0}".format(str(err)))
    if args.generic_url:
        print(rel.https_url_generic)
        sys.exit()
    imgs = rel.all_images

    # arch
    if args.arch:
        if 'i386' in args.arch and not 'i686' in args.arch:
            args.arch.append('i686')
        elif 'i686' in args.arch and not 'i386' in args.arch:
            args.arch.append('i386')
        imgs = [img for img in imgs if img['arch'].lower() in args.arch]

    # type
    if args.type:
        # we never use the 'netinst' type any more, so be nice and
        # replace it with 'boot'
        if 'netinst' in args.type:
            logger.warning('Type "netinst" will never match, since fedfind 3.1.0! Using type '
                           '"boot" instead.')
            args.type.remove('netinst')
            args.type.append('boot')
        imgs = [img for img in imgs if img['type'].lower() in args.type]

    # payload
    if args.payload:
        imgs = [img for img in imgs if img['subvariant'].lower() in args.payload]

    # search
    if args.search:
        imgs = [img for img in imgs if any(term in img['path'].lower() for term in args.search)]

    for img in imgs:
        logger.debug("Subvariant: %s Type: %s Arch: %s", img['subvariant'],
                     img['type'], img['arch'])
        print(img['url'])

def parse_args():
    """Parse arguments with argparse."""
    parser = argparse.ArgumentParser(description=(
        "Tool for finding Fedora stuff. Currently finds images, using the "
        "images sub-command. See the image help for more details."))
    parser.add_argument(
        '-l', '--loglevel', help="The level of log messages to show",
        choices=('debug', 'info', 'warning', 'error', 'critical'), default='info')
    # This is a workaround for a somewhat infamous argparse bug
    # in Python 3. See:
    # https://stackoverflow.com/questions/23349349/argparse-with-required-subparser
    # http://bugs.python.org/issue16308
    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = True
    parser_images = subparsers.add_parser(
        'images', description="Find Fedora images. You can specify --composeid, --release and "
        "--label (and --dist if necessary), or --release, --milestone, --compose and --respin (and "
        "--dist if necessary) to identify the release you wish to find images for: e.g. -r 24 -m "
        "Beta -c 1 -i 1, -r 24 -m Branched -c 20160319 -i 0, -r 24 -m Alpha, or -r 22. If the "
        "values supplied do not entirely define a release, fedfind will usually try to guess what "
        "you mean. The other parameters can be used to filter the results in various ways. For "
        "all of the query parameters, you can specify a single value or several separated by "
        "commas. If you specify several values, fedfind will find images that match *any* value. "
        "If you pass multiple query parameters, fedfind will only find images that pass the check "
        "for all of the parameters. Matching is exact (but not case-sensitive) for all parameters "
        "except --search, which will match any image where at least one of the search terms occurs "
        "somewhere in the image URL.")
    parser_images.add_argument(
        '-d', '--dist', help="The 'dist' to search for. For most composes, this will be 'Fedora', "
        "the default value. To find two-week Atomic nightly composes, use 'Fedora-Atomic'. For "
        "post-release nightly Docker composes, use 'Fedora-Docker'. For post-release nightly Cloud "
        "composes, use 'Fedora-Cloud'. For the unofficial post-release live respin composes, use "
        "'FedoraRespin'.",
        choices=['Fedora', 'Fedora-Atomic', 'Fedora-Docker', 'Fedora-Cloud', 'FedoraRespin'],
        default='Fedora', metavar="Fedora, Fedora-Atomic...")
    parser_images.add_argument(
        '-r', '--release', help="The Fedora release to search",
        type=_release, required=False, metavar="1-99 or Rawhide")
    parser_images.add_argument(
        '-m', '--milestone', help="A milestone to search (e.g. Alpha or Beta). 'Final' and 'RC' "
        "are synonyms. 'Branched' and 'Rawhide' are the two nightly compose types for 'Fedora' "
        "composes, see the wiki for more details. 'Production' is for candidate composes on the "
        "compose server (kojipkgs) and identified by a date and respin, as opposed to the same "
        "composes synced to alt and identified by label.", choices=
        ['Alpha', 'Beta', 'Final', 'RC', 'Atomic', 'Docker', 'Cloud', 'Branched', 'Rawhide',
         'Production', 'Respin'])
    parser_images.add_argument(
        '-c', '--compose', '--date', help="A compose or date to search, e.g. 1 or 20160314. You "
        "may also pass a compose and respin combined, e.g. 1.2", required=False,
        metavar="20150213, 1, 1.1")
    parser_images.add_argument('-i', '--respin', required=False, type=int,
                               help="The respin of the compose to search (an integer)")
    parser_images.add_argument(
        '--composeid', help="A compose ID to search for. This is an alternative to release/"
        "milestone/compose/respin(/dist) for identifying the compose you want. compose ID format "
        "is DIST-REL-YYYYMMDD(.T).N, where DIST is the 'dist' or 'shortname' (usually 'Fedora'), "
        "REL is e.g. Rawhide or 24, T is a type identifier - 'n' for nightly, 't' for test (not "
        "used by Fedora yet), or omitted for production - and N is the respin. e.g. "
        "Fedora-Rawhide-20160319.n.0 or Fedora-24-20160319.1")
    parser_images.add_argument(
        '-l', '--label', help="A compose label to search for. Only Pungi 4 production composes "
        "have labels. Label format is MILESTONE-C.R, where MILESTONE is e.g. Alpha or Beta, C is "
        "an integer compose number, and R is an integer respin number, e.g. Alpha-1.5. You should "
        "also pass --release to identify the release.")
    parser_images.add_argument(
        '-a', '--arch', help="Architecture(s) to search for", required=False,
        type=fedfind.helpers.comma_list, metavar="armhfp,x86_64...")
    parser_images.add_argument(
        '-t', '--type', help="Image type(s) to search for", required=False,
        metavar="boot,dvd,live...", type=fedfind.helpers.comma_list)
    parser_images.add_argument(
        '-p', '--payload', help="Image payload (subvariant) to search for", required=False,
        metavar="workstation,lxde...", type=fedfind.helpers.comma_list)
    parser_images.add_argument(
        '-s', '--search', help="String(s) to search for anywhere in image URL", required=False,
        type=fedfind.helpers.comma_list, metavar="TERM1,TERM2")
    parser_images.add_argument(
        '-g', '--generic-url', help="Just display the HTTPS URL for the 'generic' tree for the "
        "given release - the preferred source for PXE boot kernel/initramfs and so on.",
        required=False, action='store_true')
    parser_images.set_defaults(func=images)
    return parser.parse_args()

def run():
    """Read in arguments, set up logging and run sub-command
    function."""
    args = parse_args()
    loglevel = getattr(logging, args.loglevel.upper(), logging.INFO)
    logging.basicConfig(level=loglevel)
    args.func(args)

def main():
    """Main loop."""
    try:
        run()
    except KeyboardInterrupt:
        sys.stderr.write("Interrupted, exiting...\n")
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) Red Hat Inc.
#
# fedfind is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author:   Adam Williamson <awilliam@redhat.com>

"""Library and CLI tool for locating and getting information about
Fedora composes.
"""

from __future__ import unicode_literals
from __future__ import print_function

__version__ = "4.0.0"

# vim: set textwidth=100 ts=8 et sw=4:
